"""
JIRA adapter module for the document pointer system.

This module provides an adapter to retrieve content from Atlassian JIRA sources.
"""

import json
import logging
from datetime import datetime
from typing import Dict, Any, Optional

from .base import ContentSourceAdapter
from ..document_parser.document_type_detector import DocumentTypeDetector

logger = logging.getLogger(__name__)

# Try to import requests, but don't fail if not available
try:
    import requests
    from bs4 import BeautifulSoup

    REQUESTS_AVAILABLE = True
except ImportError:
    requests = None
    BeautifulSoup = None
    REQUESTS_AVAILABLE = False
    logger.warning("requests is required for JIRA adapter. Install with 'pip install requests beautifulsoup4'")


class JiraAdapter(ContentSourceAdapter):
    """Adapter for Atlassian JIRA content."""

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """Initialize the JIRA adapter."""
        if not REQUESTS_AVAILABLE:
            raise ImportError("requests and beautifulsoup4 are required for JIRA adapter")

        super().__init__(config)
        self.config = config or {}

        # JIRA connection settings
        self.base_url = self.config.get("base_url")
        self.api_path = "/rest/api/2"
        self.timeout = self.config.get("timeout", 30)
        self.verify_ssl = self.config.get("verify_ssl", True)

        # Authentication settings
        self.auth_type = self.config.get("auth_type", "basic")
        self.username = self.config.get("username")
        self.password = self.config.get("password")
        self.api_token = self.config.get("api_token")

        # Optional settings
        self.include_attachments = self.config.get("include_attachments", False)
        self.include_comments = self.config.get("include_comments", True)
        self.include_changelog = self.config.get("include_changelog", False)
        self.max_results = self.config.get("max_results", 100)
        self.fields = self.config.get("fields", "*all")

        # Initialize session
        self.sessions = {}

        # Caches
        self.content_cache = {}
        self.binary_cache = {}
        self.metadata_cache = {}
        self.field_cache = {}
        self.project_cache = {}

    def get_content(self, location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get content from JIRA.

        Args:
            location_data: Location data with JIRA information

        Returns:
            Dictionary with content and metadata

        Raises:
            ValueError: If JIRA content cannot be retrieved
        """
        source = location_data.get("source", "")
        if not source.startswith("jira://"):
            raise ValueError(f"Invalid JIRA source: {source}")

        # Check cache first
        if source in self.content_cache:
            return self.content_cache[source]

        # Parse JIRA URI
        parsed_data = self._parse_jira_uri(source)

        # Get session for this base URL
        session = self._get_session(parsed_data["base_url"])

        try:
            # Determine the type of resource we need to fetch
            content_type = parsed_data.get("content_type", "issue")

            # Handle different content types
            if content_type == "issue":
                return self._get_issue_content(session, parsed_data, location_data)
            elif content_type == "attachment":
                return self._get_attachment_content(session, parsed_data, location_data)
            elif content_type == "comment":
                return self._get_comment_content(session, parsed_data, location_data)
            elif content_type == "project":
                return self._get_project_content(session, parsed_data, location_data)
            elif content_type == "search":
                return self._get_search_content(session, parsed_data, location_data)
            elif content_type == "field":
                return self._get_field_content(session, parsed_data, location_data)
            else:
                raise ValueError(f"Unsupported JIRA content type: {content_type}")

        except Exception as e:
            logger.error(f"Error retrieving JIRA content: {str(e)}")
            raise ValueError(f"Error retrieving JIRA content: {str(e)}")

    def supports_location(self, location_data: Dict[str, Any]) -> bool:
        """
        Check if this adapter supports the location.

        Args:
            location_data: Content location data

        Returns:
            True if supported, False otherwise
        """
        if not REQUESTS_AVAILABLE:
            return False

        source = location_data.get("source", "")
        # Source must be a JIRA URI
        return source.startswith("jira://")

    def get_binary_content(self, location_data: Dict[str, Any]) -> bytes:
        """
        Get the JIRA content as binary data.

        Args:
            location_data: Location data

        Returns:
            Binary content

        Raises:
            ValueError: If content cannot be retrieved as binary
        """
        source = location_data.get("source", "")

        # Check if it's specifically an attachment
        parsed_data = self._parse_jira_uri(source)

        # If it's an attachment, get the binary content directly
        if parsed_data.get("content_type") == "attachment":
            # Check cache first
            if source in self.binary_cache:
                return self.binary_cache[source]

            # Get session
            session = self._get_session(parsed_data["base_url"])

            try:
                # Construct URL
                attachment_id = parsed_data.get("attachment_id")
                # issue_key = parsed_data.get("issue_key")

                # Fetch attachment metadata first to get the content URL
                metadata_url = f"{parsed_data['base_url']}{self.api_path}/attachment/{attachment_id}"

                response = session.get(
                    metadata_url,
                    timeout=self.timeout,
                    verify=self.verify_ssl
                )
                response.raise_for_status()

                # Parse response
                attachment_data = response.json()

                # Get download URL
                download_url = attachment_data.get("content")

                if not download_url:
                    raise ValueError(f"Could not find download URL for attachment {attachment_id}")

                # Fetch attachment
                response = session.get(
                    download_url,
                    timeout=self.timeout,
                    verify=self.verify_ssl
                )
                response.raise_for_status()

                # Get binary content
                binary_data = response.content

                # Cache the result
                self.binary_cache[source] = binary_data

                return binary_data

            except Exception as e:
                logger.error(f"Error retrieving attachment binary: {str(e)}")
                raise ValueError(f"Error retrieving attachment binary: {str(e)}")

        # For non-attachments, convert the content to bytes
        content_info = self.get_content(location_data)
        content = content_info.get("content", "")

        # If content is already bytes, return it
        if isinstance(content, (bytes, bytearray)):
            return content

        # Convert string content to bytes
        if isinstance(content, str):
            return content.encode('utf-8')

        # For other types, convert to JSON and then to bytes
        if isinstance(content, (dict, list)):
            return json.dumps(content).encode('utf-8')

        # As a last resort, convert to string
        return str(content).encode('utf-8')

    def get_metadata(self, location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get metadata about the JIRA content without retrieving the full content.

        Args:
            location_data: Location data

        Returns:
            Dictionary with metadata

        Raises:
            ValueError: If metadata cannot be retrieved
        """
        source = location_data.get("source", "")

        # Check cache first
        if source in self.metadata_cache:
            return self.metadata_cache[source]

        # Parse JIRA URI
        parsed_data = self._parse_jira_uri(source)

        # Get session for this base URL
        session = self._get_session(parsed_data["base_url"])

        try:
            # Determine content type
            content_type = parsed_data.get("content_type", "issue")

            # Base API URL
            api_url = f"{parsed_data['base_url']}{self.api_path}"

            # Prepare metadata dictionary
            metadata = {
                "base_url": parsed_data["base_url"],
                "content_type": content_type,
                "retrieved_at": datetime.now().isoformat()
            }

            # Handle different content types
            if content_type == "issue":
                # Fetch issue metadata
                issue_key = parsed_data.get("issue_key")

                if issue_key:
                    # Construct API URL - just get fields, not content
                    issue_url = f"{api_url}/issue/{issue_key}"
                    params = {
                        "fields": "summary,issuetype,status,created,updated,project,creator,reporter"
                    }

                    # Fetch issue
                    response = session.get(
                        issue_url,
                        params=params,
                        timeout=self.timeout,
                        verify=self.verify_ssl
                    )
                    response.raise_for_status()

                    # Parse response
                    issue_data = response.json()
                    fields = issue_data.get("fields", {})

                    # Extract metadata
                    metadata.update({
                        "id": issue_data.get("id"),
                        "key": issue_data.get("key"),
                        "summary": fields.get("summary"),
                        "issue_type": fields.get("issuetype", {}).get("name"),
                        "status": fields.get("status", {}).get("name"),
                        "created_at": fields.get("created"),
                        "updated_at": fields.get("updated"),
                        "project_key": fields.get("project", {}).get("key"),
                        "project_name": fields.get("project", {}).get("name"),
                        "creator": fields.get("creator", {}).get("displayName"),
                        "reporter": fields.get("reporter", {}).get("displayName")
                    })

            elif content_type == "attachment":
                # Fetch attachment metadata
                attachment_id = parsed_data.get("attachment_id")

                if attachment_id:
                    # Construct API URL
                    attachment_url = f"{api_url}/attachment/{attachment_id}"

                    # Fetch metadata
                    response = session.get(
                        attachment_url,
                        timeout=self.timeout,
                        verify=self.verify_ssl
                    )
                    response.raise_for_status()

                    # Parse response
                    attachment_data = response.json()

                    # Extract metadata
                    metadata.update({
                        "id": attachment_data.get("id"),
                        "filename": attachment_data.get("filename"),
                        "created_at": attachment_data.get("created"),
                        "author": attachment_data.get("author", {}).get("displayName"),
                        "size": attachment_data.get("size"),
                        "mime_type": attachment_data.get("mimeType"),
                        "issue_id": attachment_data.get("self", "").split("/")[-3],
                        "issue_key": parsed_data.get("issue_key")
                    })

            elif content_type == "comment":
                # Fetch comment metadata
                comment_id = parsed_data.get("comment_id")
                issue_key = parsed_data.get("issue_key")

                if comment_id and issue_key:
                    # Construct API URL
                    comment_url = f"{api_url}/issue/{issue_key}/comment/{comment_id}"

                    # Fetch comment
                    response = session.get(
                        comment_url,
                        timeout=self.timeout,
                        verify=self.verify_ssl
                    )
                    response.raise_for_status()

                    # Parse response
                    comment_data = response.json()

                    # Extract metadata
                    metadata.update({
                        "id": comment_data.get("id"),
                        "created_at": comment_data.get("created"),
                        "updated_at": comment_data.get("updated"),
                        "author": comment_data.get("author", {}).get("displayName"),
                        "issue_key": issue_key
                    })

            elif content_type == "project":
                # Fetch project metadata
                project_key = parsed_data.get("project_key")

                if project_key:
                    # Construct API URL
                    project_url = f"{api_url}/project/{project_key}"

                    # Fetch project
                    response = session.get(
                        project_url,
                        timeout=self.timeout,
                        verify=self.verify_ssl
                    )
                    response.raise_for_status()

                    # Parse response
                    project_data = response.json()

                    # Extract metadata
                    metadata.update({
                        "id": project_data.get("id"),
                        "key": project_data.get("key"),
                        "name": project_data.get("name"),
                        "project_type": project_data.get("projectTypeKey"),
                        "lead": project_data.get("lead", {}).get("displayName")
                    })

            # Cache the metadata
            self.metadata_cache[source] = metadata

            return metadata

        except Exception as e:
            logger.error(f"Error retrieving JIRA metadata: {str(e)}")
            raise ValueError(f"Error retrieving JIRA metadata: {str(e)}")

    def resolve_uri(self, uri: str) -> Dict[str, Any]:
        """
        Parse a JIRA URI into location data.

        Args:
            uri: JIRA URI string

        Returns:
            Dictionary with parsed location data

        Raises:
            ValueError: If URI cannot be parsed
        """
        if not uri.startswith("jira://"):
            raise ValueError(f"Not a JIRA URI: {uri}")

        # Parse URI into components
        parsed_data = self._parse_jira_uri(uri)

        # Build location data
        location_data = {
            "source": uri
        }

        # Add all components to the location data
        location_data.update(parsed_data)

        return location_data

    def cleanup(self):
        """
        Clean up resources used by this adapter.

        This method should be called when the adapter is no longer needed.
        """
        # Close sessions
        for base_url, session in self.sessions.items():
            try:
                session.close()
                logger.debug(f"Closed JIRA session for: {base_url}")
            except Exception as e:
                logger.warning(f"Error closing JIRA session for {base_url}: {str(e)}")

        # Clear caches and session
        self.sessions = {}
        self.content_cache = {}
        self.binary_cache = {}
        self.metadata_cache = {}
        self.field_cache = {}
        self.project_cache = {}

    @staticmethod
    def _parse_jira_uri(uri: str) -> Dict[str, Any]:
        """
        Parse JIRA URI into components.

        Expected formats:
        - jira://domain/issue/{issue_key}
        - jira://domain/issue/{issue_key}/{field_id}
        - jira://domain/issue/{issue_key}/attachment/{attachment_id}
        - jira://domain/issue/{issue_key}/comment/{comment_id}
        - jira://domain/project/{project_key}
        - jira://domain/search/{jql_query}
        - jira://domain/field/{field_id}

        Args:
            uri: JIRA URI string

        Returns:
            Dictionary with parsed components
        """
        if not uri.startswith("jira://"):
            raise ValueError(f"Invalid JIRA URI: {uri}")

        # Remove the jira:// prefix
        path = uri[7:]

        # Split path
        parts = path.split('/')

        # We need at least a domain
        if len(parts) < 1:
            raise ValueError(f"Invalid JIRA URI format: {uri}")

        # Extract domain and build base URL
        domain = parts[0]
        if not domain.startswith(('http://', 'https://')):
            base_url = f"https://{domain}"
        else:
            base_url = domain

        # Create result with base URL
        result = {
            "base_url": base_url,
            "source": uri
        }

        # If we only have the domain, assume default content type
        if len(parts) == 1:
            result["content_type"] = "issue"
            return result

        # Check for content type
        if len(parts) >= 2:
            content_type = parts[1]
            result["content_type"] = content_type

            # Handle different content types
            if content_type == "issue" and len(parts) >= 3:
                # Issue URI - always include issue key
                result["issue_key"] = parts[2]

                # Check for additional components
                if len(parts) >= 5 and parts[3] == "attachment":
                    # Attachment URI
                    result["content_type"] = "attachment"
                    result["attachment_id"] = parts[4]
                elif len(parts) >= 5 and parts[3] == "comment":
                    # Comment URI
                    result["content_type"] = "comment"
                    result["comment_id"] = parts[4]
                elif len(parts) >= 4:
                    # Field URI
                    result["field_id"] = parts[3]

            elif content_type == "project" and len(parts) >= 3:
                # Project URI
                result["project_key"] = parts[2]

            elif content_type == "search" and len(parts) >= 3:
                # Search URI - extract JQL
                result["jql"] = "/".join(parts[2:])

            elif content_type == "field" and len(parts) >= 3:
                # Field URI
                result["field_id"] = parts[2]

        return result

    def _get_session(self, base_url: str) -> requests.Session:
        """
        Get or create a session for the given base URL.

        Args:
            base_url: JIRA base URL

        Returns:
            Requests session with authentication
        """
        # Check if session exists in cache
        if base_url in self.sessions:
            return self.sessions[base_url]

        # Create new session
        session = requests.Session()

        # Add authentication
        if self.auth_type == "basic":
            if self.username and (self.password or self.api_token):
                # Use API token if available, otherwise use password
                auth_password = self.api_token or self.password
                session.auth = (self.username, auth_password)
        elif self.auth_type == "bearer":
            if self.api_token:
                session.headers.update({
                    "Authorization": f"Bearer {self.api_token}"
                })

        # Add common headers
        session.headers.update({
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Atlassian-Token": "no-check"
        })

        # Cache session
        self.sessions[base_url] = session

        return session

    def _get_issue_content(self, session: requests.Session, parsed_data: Dict[str, Any],
                           _location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get JIRA issue content.

        Args:
            session: Requests session
            parsed_data: Parsed JIRA URI data
            _location_data: Original location data

        Returns:
            Dictionary with content and metadata
        """
        base_url = parsed_data["base_url"]
        issue_key = parsed_data.get("issue_key")
        field_id = parsed_data.get("field_id")

        if not issue_key:
            raise ValueError("Issue key is required for issue content")

        # Construct API URL
        api_url = f"{base_url}{self.api_path}/issue/{issue_key}"

        # Determine what to expand
        expand = []

        if self.include_attachments:
            expand.append("attachment")

        if self.include_comments:
            expand.append("comment")

        if self.include_changelog:
            expand.append("changelog")

        params = {
            "fields": self.fields
        }

        if expand:
            params["expand"] = ",".join(expand)

        # Fetch issue
        response = session.get(
            api_url,
            params=params,
            timeout=self.timeout,
            verify=self.verify_ssl
        )
        response.raise_for_status()

        # Parse response
        issue_data = response.json()

        # Determine what content to return based on field_id
        # content = None
        content_type = "json"

        if field_id:
            # Get specific field
            fields = issue_data.get("fields", {})

            # Check if this is a custom field ID
            if field_id.startswith("customfield_"):
                # Direct access to custom field
                content = fields.get(field_id)
            else:
                # Map field ID to field name
                field_map = self._get_field_map(session, base_url)
                actual_field_id = field_map.get(field_id, field_id)
                content = fields.get(actual_field_id)

            # If field value is complex, return as JSON
            if isinstance(content, (dict, list)):
                content_type = "json"
                # Convert to string for consistency
                content = json.dumps(content, indent=2)
            else:
                # For simple values, treat as text
                content_type = "text"
                content = str(content) if content is not None else ""

        else:
            # Return entire issue as JSON
            content = json.dumps(issue_data, indent=2)

        # Extract metadata
        fields = issue_data.get("fields", {})
        metadata = {
            "id": issue_data.get("id"),
            "key": issue_data.get("key"),
            "url": f"{base_url}/browse/{issue_data.get('key')}",
            "summary": fields.get("summary"),
            "issue_type": fields.get("issuetype", {}).get("name"),
            "status": fields.get("status", {}).get("name"),
            "created_at": fields.get("created"),
            "updated_at": fields.get("updated"),
            "project_key": fields.get("project", {}).get("key"),
            "project_name": fields.get("project", {}).get("name"),
            "creator": fields.get("creator", {}).get("displayName"),
            "reporter": fields.get("reporter", {}).get("displayName"),
            "base_url": base_url,
            "content_type": "issue"
        }

        # Add attachment info if available
        attachments = fields.get("attachment", [])
        if attachments:
            metadata["attachment_count"] = len(attachments)
            metadata["attachments"] = [{
                "id": att.get("id"),
                "filename": att.get("filename"),
                "size": att.get("size")
            } for att in attachments]

        # Add comment info if available
        comments = fields.get("comment", {}).get("comments", [])
        if comments:
            metadata["comment_count"] = len(comments)

        # Create result
        result = {
            "content": content,
            "content_type": content_type,
            "metadata": metadata
        }

        # Cache the result
        self.content_cache[parsed_data["source"]] = result
        self.metadata_cache[parsed_data["source"]] = metadata

        return result

    def _get_attachment_content(self, session: requests.Session, parsed_data: Dict[str, Any],
                                _location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get JIRA attachment content.

        Args:
            session: Requests session
            parsed_data: Parsed JIRA URI data
            _location_data: Original location data

        Returns:
            Dictionary with content and metadata
        """
        base_url = parsed_data["base_url"]
        issue_key = parsed_data.get("issue_key")
        attachment_id = parsed_data.get("attachment_id")

        if not attachment_id:
            raise ValueError("Attachment ID is required for attachment content")

        # Construct API URL for attachment metadata
        api_url = f"{base_url}{self.api_path}/attachment/{attachment_id}"

        # Fetch attachment metadata
        response = session.get(
            api_url,
            timeout=self.timeout,
            verify=self.verify_ssl
        )
        response.raise_for_status()

        # Parse response
        attachment_data = response.json()

        # Get download URL
        download_url = attachment_data.get("content")

        if not download_url:
            raise ValueError(f"Could not find download URL for attachment {attachment_id}")

        # Fetch attachment content
        response = session.get(
            download_url,
            timeout=self.timeout,
            verify=self.verify_ssl
        )
        response.raise_for_status()

        # Get content
        content = response.content

        # Determine content type from MIME type
        mime_type = attachment_data.get("mimeType", "")
        filename = attachment_data.get("filename", "")

        # Detect content type
        if mime_type:
            content_type = self._mime_type_to_content_type(mime_type, filename)
        else:
            content_type = DocumentTypeDetector.detect(
                path=filename,
                content=content,
                metadata={"binary": True}
            )

        # Try to convert to text if it's a text-based format
        if self._is_text_content(mime_type):
            try:
                content = content.decode('utf-8')
            except UnicodeDecodeError:
                # Keep as binary if decoding fails
                pass

        # Extract metadata
        metadata = {
            "id": attachment_data.get("id"),
            "filename": attachment_data.get("filename"),
            "created_at": attachment_data.get("created"),
            "author": attachment_data.get("author", {}).get("displayName"),
            "size": attachment_data.get("size"),
            "mime_type": attachment_data.get("mimeType"),
            "download_url": download_url,
            "base_url": base_url,
            "content_type": "attachment",
            "issue_key": issue_key,
            "issue_id": attachment_data.get("self", "").split("/")[-3]
        }

        # Create result
        result = {
            "content": content,
            "content_type": content_type,
            "metadata": metadata
        }

        # Cache the result
        self.content_cache[parsed_data["source"]] = result
        self.metadata_cache[parsed_data["source"]] = metadata

        return result

    def _get_comment_content(self, session: requests.Session, parsed_data: Dict[str, Any],
                             _location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get JIRA comment content.

        Args:
            session: Requests session
            parsed_data: Parsed JIRA URI data
            _location_data: Original location data

        Returns:
            Dictionary with content and metadata
        """
        base_url = parsed_data["base_url"]
        issue_key = parsed_data.get("issue_key")
        comment_id = parsed_data.get("comment_id")

        if not comment_id or not issue_key:
            raise ValueError("Issue key and comment ID are required for comment content")

        # Construct API URL
        api_url = f"{base_url}{self.api_path}/issue/{issue_key}/comment/{comment_id}"

        # Fetch comment
        response = session.get(
            api_url,
            timeout=self.timeout,
            verify=self.verify_ssl
        )
        response.raise_for_status()

        # Parse response
        comment_data = response.json()

        # Get comment body in different formats
        content = comment_data.get("body", "")

        # Try to extract rendered content if available
        rendered_content = comment_data.get("renderedBody", "")
        if rendered_content:
            content = rendered_content
            content_type = "html"
        else:
            # Use plain text
            content_type = "text"

        # Extract metadata
        metadata = {
            "id": comment_data.get("id"),
            "created_at": comment_data.get("created"),
            "updated_at": comment_data.get("updated"),
            "author": comment_data.get("author", {}).get("displayName"),
            "base_url": base_url,
            "content_type": "comment",
            "issue_key": issue_key
        }

        # Create result
        result = {
            "content": content,
            "content_type": content_type,
            "metadata": metadata
        }

        # Cache the result
        self.content_cache[parsed_data["source"]] = result
        self.metadata_cache[parsed_data["source"]] = metadata

        return result

    def _get_project_content(self, session: requests.Session, parsed_data: Dict[str, Any],
                             _location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get JIRA project content.

        Args:
            session: Requests session
            parsed_data: Parsed JIRA URI data
            _location_data: Original location data

        Returns:
            Dictionary with content and metadata
        """
        base_url = parsed_data["base_url"]
        project_key = parsed_data.get("project_key")

        if not project_key:
            raise ValueError("Project key is required for project content")

        # Construct API URL
        api_url = f"{base_url}{self.api_path}/project/{project_key}"

        # Fetch project
        response = session.get(
            api_url,
            timeout=self.timeout,
            verify=self.verify_ssl
        )
        response.raise_for_status()

        # Parse response
        project_data = response.json()

        # Get project content as JSON
        content = json.dumps(project_data, indent=2)
        content_type = "json"

        # Extract metadata
        metadata = {
            "id": project_data.get("id"),
            "key": project_data.get("key"),
            "name": project_data.get("name"),
            "url": f"{base_url}/projects/{project_data.get('key')}",
            "project_type": project_data.get("projectTypeKey"),
            "lead": project_data.get("lead", {}).get("displayName"),
            "base_url": base_url,
            "content_type": "project"
        }

        # Create result
        result = {
            "content": content,
            "content_type": content_type,
            "metadata": metadata
        }

        # Cache the result
        self.content_cache[parsed_data["source"]] = result
        self.metadata_cache[parsed_data["source"]] = metadata
        self.project_cache[project_key] = project_data

        return result

    def _get_search_content(self, session: requests.Session, parsed_data: Dict[str, Any],
                            _location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get JIRA search content.

        Args:
            session: Requests session
            parsed_data: Parsed JIRA URI data
            _location_data: Original location data

        Returns:
            Dictionary with content and metadata
        """
        base_url = parsed_data["base_url"]
        jql = parsed_data.get("jql", "")

        if not jql:
            raise ValueError("JQL query is required for search content")

        # Construct API URL
        api_url = f"{base_url}{self.api_path}/search"

        # Prepare params
        params = {
            "jql": jql,
            "maxResults": self.max_results,
            "fields": self.fields
        }

        # Fetch search results
        response = session.get(
            api_url,
            params=params,
            timeout=self.timeout,
            verify=self.verify_ssl
        )
        response.raise_for_status()

        # Parse response
        search_data = response.json()

        # Get search content as JSON
        content = json.dumps(search_data, indent=2)
        content_type = "json"

        # Extract metadata
        metadata = {
            "jql": jql,
            "total": search_data.get("total", 0),
            "max_results": self.max_results,
            "start_at": search_data.get("startAt", 0),
            "base_url": base_url,
            "content_type": "search"
        }

        # Create result
        result = {
            "content": content,
            "content_type": content_type,
            "metadata": metadata
        }

        # Cache the result
        self.content_cache[parsed_data["source"]] = result
        self.metadata_cache[parsed_data["source"]] = metadata

        return result

    def _get_field_content(self, session: requests.Session, parsed_data: Dict[str, Any],
                           _location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get JIRA field content.

        Args:
            session: Requests session
            parsed_data: Parsed JIRA URI data
            _location_data: Original location data

        Returns:
            Dictionary with content and metadata
        """
        base_url = parsed_data["base_url"]
        field_id = parsed_data.get("field_id")

        if not field_id:
            # Get all fields
            api_url = f"{base_url}{self.api_path}/field"

            # Fetch fields
            response = session.get(
                api_url,
                timeout=self.timeout,
                verify=self.verify_ssl
            )
            response.raise_for_status()

            # Parse response
            fields_data = response.json()

            # Get fields content as JSON
            content = json.dumps(fields_data, indent=2)
            content_type = "json"

            # Extract metadata
            metadata = {
                "field_count": len(fields_data),
                "base_url": base_url,
                "content_type": "field"
            }

            # Create result
            result = {
                "content": content,
                "content_type": content_type,
                "metadata": metadata
            }

            # Cache the result
            self.content_cache[parsed_data["source"]] = result
            self.metadata_cache[parsed_data["source"]] = metadata

            return result
        else:
            # Get specific field
            field_map = self._get_field_map(session, base_url)

            # Find field by ID or name
            field_data = None

            # Check if field exists in the map
            if field_id in field_map:
                field_data = field_map[field_id]
            else:
                # Get all fields
                api_url = f"{base_url}{self.api_path}/field"

                # Fetch fields
                response = session.get(
                    api_url,
                    timeout=self.timeout,
                    verify=self.verify_ssl
                )
                response.raise_for_status()

                # Parse response
                fields_data = response.json()

                # Find field
                for field in fields_data:
                    if field.get("id") == field_id or field.get("name") == field_id:
                        field_data = field
                        break

            if not field_data:
                raise ValueError(f"Field not found: {field_id}")

            # Get field content as JSON
            content = json.dumps(field_data, indent=2)
            content_type = "json"

            # Extract metadata
            metadata = {
                "id": field_data.get("id"),
                "name": field_data.get("name"),
                "custom": field_data.get("custom", False),
                "schema": field_data.get("schema"),
                "base_url": base_url,
                "content_type": "field"
            }

            # Create result
            result = {
                "content": content,
                "content_type": content_type,
                "metadata": metadata
            }

            # Cache the result
            self.content_cache[parsed_data["source"]] = result
            self.metadata_cache[parsed_data["source"]] = metadata

            return result

    def _get_field_map(self, session: requests.Session, base_url: str) -> Dict[str, Any]:
        """
        Get field name to ID mapping.

        Args:
            session: Requests session
            base_url: JIRA base URL

        Returns:
            Dictionary mapping field names to field IDs
        """
        # Check cache first
        cache_key = f"field_map_{base_url}"
        if cache_key in self.field_cache:
            return self.field_cache[cache_key]

        # Construct API URL
        api_url = f"{base_url}{self.api_path}/field"

        try:
            # Fetch fields
            response = session.get(
                api_url,
                timeout=self.timeout,
                verify=self.verify_ssl
            )
            response.raise_for_status()

            # Parse response
            fields_data = response.json()

            # Build mapping
            field_map = {}

            for field in fields_data:
                field_id = field.get("id")
                field_name = field.get("name")
                if field_id and field_name:
                    field_map[field_name] = field_id
                    field_map[field_id] = field

            # Cache the mapping
            self.field_cache[cache_key] = field_map

            return field_map

        except Exception as e:
            logger.error(f"Error fetching JIRA fields: {str(e)}")
            return {}

    @staticmethod
    def _mime_type_to_content_type(mime_type: str, filename: str) -> str:
        """
        Convert MIME type to content type.

        Args:
            mime_type: MIME type
            filename: Filename for additional context

        Returns:
            Content type string
        """
        # Extract extension from filename
        extension = ""
        if filename and "." in filename:
            extension = filename.split(".")[-1].lower()

        # Handle common document types
        if mime_type.startswith("text/html") or extension in ["html", "htm"]:
            return "html"
        elif mime_type.startswith("text/markdown") or extension in ["md", "markdown"]:
            return "markdown"
        elif mime_type.startswith("application/pdf") or extension == "pdf":
            return "pdf"
        elif mime_type.startswith(
                "application/vnd.openxmlformats-officedocument.wordprocessingml.document") or extension == "docx":
            return "docx"
        elif mime_type.startswith(
                "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet") or extension == "xlsx":
            return "xlsx"
        elif mime_type.startswith(
                "application/vnd.openxmlformats-officedocument.presentationml.presentation") or extension == "pptx":
            return "pptx"
        elif mime_type.startswith("application/json") or extension == "json":
            return "json"
        elif mime_type.startswith("application/xml") or extension in ["xml", "svg"]:
            return "xml"
        elif mime_type.startswith("text/plain") or extension == "txt":
            return "text"
        elif mime_type.startswith("text/csv") or extension == "csv":
            return "csv"
        elif mime_type.startswith("image/"):
            return "image"
        elif mime_type.startswith("video/"):
            return "video"
        elif mime_type.startswith("audio/"):
            return "audio"

        # Default to binary for unknown types
        return "binary"

    @staticmethod
    def _is_text_content(mime_type: str) -> bool:
        """
        Determine if a MIME type represents text-based content.

        Args:
            mime_type: MIME type

        Returns:
            True if text-based, False otherwise
        """
        if not mime_type:
            return False

        # Text MIME types
        text_types = [
            'text/',
            'application/json',
            'application/xml',
            'application/yaml',
            'application/x-yaml',
            'application/javascript',
            'application/typescript',
            'application/csv',
            'application/x-csv',
            'application/markdown',
            'application/x-markdown'
        ]

        # Check if MIME type starts with any of the text types
        for text_type in text_types:
            if mime_type.startswith(text_type):
                return True

        return False



"""Automatically generated __init__.py"""
__all__ = ['DateTimeEncoder', 'DocumentDatabase', 'ElementBase', 'ElementFlat', 'ElementHierarchical',
           'ElementRelationship', 'ElementType', 'FileDocumentDatabase', 'MongoDBDocumentDatabase',
           'Neo4jDocumentDatabase', 'PostgreSQLDocumentDatabase', 'RelationshipCategory', 'SQLAlchemyDocumentDatabase',
           'SQLiteDocumentDatabase', 'base', 'build_element_hierarchy', 'element_element', 'element_relationship',
           'factory', 'file', 'filter_elements_by_type', 'flatten_hierarchy', 'get_child_elements',
           'get_container_elements', 'get_container_relationships', 'get_document_database', 'get_explicit_links',
           'get_leaf_elements', 'get_root_elements', 'get_semantic_relationships', 'get_sibling_relationships',
           'get_structural_relationships', 'mongodb', 'neo4j_graph', 'postgres', 'sort_relationships_by_confidence',
           'sort_semantic_relationships_by_similarity', 'sqlalchemy', 'sqlite']

from . import base
from . import element_element
from . import element_relationship
from . import factory
from . import file
from . import mongodb
from . import neo4j_graph
from . import postgres
from . import sqlalchemy
from . import sqlite
from .base import DocumentDatabase
from .element_element import ElementBase
from .element_element import ElementFlat
from .element_element import ElementHierarchical
from .element_element import ElementType
from .element_element import build_element_hierarchy
from .element_element import filter_elements_by_type
from .element_element import flatten_hierarchy
from .element_element import get_child_elements
from .element_element import get_container_elements
from .element_element import get_leaf_elements
from .element_element import get_root_elements
from .element_relationship import ElementRelationship
from .element_relationship import RelationshipCategory
from .element_relationship import get_container_relationships
from .element_relationship import get_explicit_links
from .element_relationship import get_semantic_relationships
from .element_relationship import get_sibling_relationships
from .element_relationship import get_structural_relationships
from .element_relationship import sort_relationships_by_confidence
from .element_relationship import sort_semantic_relationships_by_similarity
from .factory import get_document_database
from .file import FileDocumentDatabase
from .mongodb import MongoDBDocumentDatabase
from .neo4j_graph import DateTimeEncoder
from .neo4j_graph import Neo4jDocumentDatabase
from .postgres import PostgreSQLDocumentDatabase
from .sqlalchemy import SQLAlchemyDocumentDatabase
from .sqlite import DateTimeEncoder
from .sqlite import SQLiteDocumentDatabase



import glob
import json
import logging
import os
from typing import Optional, Dict, Any, List, Tuple, Union

import time

from .base import DocumentDatabase
from .element_relationship import ElementRelationship

logger = logging.getLogger(__name__)

# Try to import the config the same way SQLite does
try:
    from ..config import Config

    config = Config(os.environ.get("DOCULYZER_CONFIG_PATH", "./config.yaml"))
except Exception as e:
    logger.warning(f"Error configuring File provider: {str(e)}")
    config = None


class FileDocumentDatabase(DocumentDatabase):
    """File-based implementation of document database."""

    def get_outgoing_relationships(self, element_pk: int) -> List[ElementRelationship]:
        """
        Find all relationships where the specified element_pk is the source.

        Implementation for File-based database, optimized to look up target elements efficiently.

        Args:
            element_pk: The primary key of the element

        Returns:
            List of ElementRelationship objects where the specified element is the source
        """
        relationships = []

        # Get the element to find its element_id and type
        element = self.get_element(element_pk)
        if not element:
            logger.warning(f"Element with PK {element_pk} not found")
            return []

        element_id = element.get("element_id")
        if not element_id:
            logger.warning(f"Element with PK {element_pk} has no element_id")
            return []

        element_type = element.get("element_type", "")

        # Create a lookup map of element_id to (element_pk, element_type, content_preview)
        # This is similar to what we get with SQL JOIN
        element_lookup = {}
        for eid, elem in self.elements.items():
            if "element_pk" in elem:
                element_lookup[eid] = (
                    elem.get("element_pk"),
                    elem.get("element_type"),
                    elem.get("content_preview", "")  # Added content_preview
                )

        # Find relationships where the element is the source (outgoing relationships)
        outgoing_relationships = [
            rel for rel in self.relationships.values()
            if rel.get("source_id") == element_id
        ]

        # Process relationships with target element lookup
        for rel in outgoing_relationships:
            target_reference = rel.get("target_reference", "")
            target_element_pk = None
            target_element_type = None
            target_content_preview = None  # Added this field

            # Look up target element information if available
            if target_reference in element_lookup:
                target_element_pk, target_element_type, target_content_preview = element_lookup[target_reference]

            # Create enriched relationship
            relationship = ElementRelationship(
                relationship_id=rel.get("relationship_id", ""),
                source_id=element_id,
                source_element_pk=element_pk,
                source_element_type=element_type,
                relationship_type=rel.get("relationship_type", ""),
                target_reference=target_reference,
                target_element_pk=target_element_pk,
                target_element_type=target_element_type,
                target_content_preview=target_content_preview,  # Added this field
                doc_id=rel.get("doc_id"),
                metadata=rel.get("metadata", {}),
                is_source=True
            )

            relationships.append(relationship)

        return relationships

    def __init__(self, storage_path: str):
        """
        Initialize file-based document database.

        Args:
            storage_path: Path to storage directory
        """
        self.storage_path = storage_path
        self.documents = {}
        self.elements = {}
        self.element_pks = {}  # Map element_id to element_pk
        self.next_element_pk = 1  # Starting auto-increment value
        self.relationships = {}
        self.embeddings = {}
        self.processing_history = {}  # Dictionary to track processing history
        self.embedding_generator = None

    def initialize(self) -> None:
        """Initialize the database by loading existing data."""
        os.makedirs(self.storage_path, exist_ok=True)

        # Create subdirectories
        os.makedirs(os.path.join(self.storage_path, 'documents'), exist_ok=True)
        os.makedirs(os.path.join(self.storage_path, 'elements'), exist_ok=True)
        os.makedirs(os.path.join(self.storage_path, 'relationships'), exist_ok=True)
        os.makedirs(os.path.join(self.storage_path, 'embeddings'), exist_ok=True)
        os.makedirs(os.path.join(self.storage_path, 'processing_history'), exist_ok=True)

        # Load existing data if available
        self._load_documents()
        self._load_elements()
        self._load_relationships()
        self._load_embeddings()
        self._load_processing_history()

        logger.info(f"Loaded {len(self.documents)} documents, "
                    f"{len(self.elements)} elements, "
                    f"{len(self.relationships)} relationships, "
                    f"{len(self.embeddings)} embeddings, "
                    f"{len(self.processing_history)} processing history records")

    def close(self) -> None:
        """Close the database (no-op for file-based database)."""
        pass

    def get_last_processed_info(self, source_id: str) -> Optional[Dict[str, Any]]:
        """Get information about when a document was last processed."""
        # Convert source_id to a safe filename
        safe_id = self._get_safe_filename(source_id)

        # Look up in processing history
        return self.processing_history.get(safe_id)

    def update_processing_history(self, source_id: str, content_hash: str) -> None:
        """Update the processing history for a document."""
        # Convert source_id to a safe filename
        safe_id = self._get_safe_filename(source_id)

        # Create history record
        history_record = {
            "source_id": source_id,
            "content_hash": content_hash,
            "last_modified": time.time(),
            "processing_count": self.processing_history.get(safe_id, {}).get("processing_count", 0) + 1
        }

        # Store in memory and on disk
        self.processing_history[safe_id] = history_record
        self._save_processing_history(safe_id)

        logger.debug(f"Updated processing history for {source_id}")

    def store_document(self, document: Dict[str, Any], elements: List[Dict[str, Any]],
                       relationships: List[Dict[str, Any]]) -> None:
        """Store a document with its elements and relationships."""
        doc_id = document["doc_id"]
        source_id = document.get("source", "")
        content_hash = document.get("content_hash", "")

        # Store document
        self.documents[doc_id] = document
        self._save_document(doc_id)

        # Store elements
        for element in elements:
            element_id = element["element_id"]

            # Assign an auto-increment element_pk
            element_pk = self.next_element_pk
            self.next_element_pk += 1
            element["element_pk"] = element_pk

            # Store mapping
            self.element_pks[element_id] = element_pk

            # Store element
            self.elements[element_id] = element
            self._save_element(element_id)

        # Store relationships
        for relationship in relationships:
            relationship_id = relationship["relationship_id"]
            self.relationships[relationship_id] = relationship
            self._save_relationship(relationship_id)

        # Update processing history
        if source_id:
            self.update_processing_history(source_id, content_hash)

    def update_document(self, doc_id: str, document: Dict[str, Any],
                        elements: List[Dict[str, Any]],
                        relationships: List[Dict[str, Any]]) -> None:
        """Update an existing document."""
        if doc_id not in self.documents:
            raise ValueError(f"Document not found: {doc_id}")

        source_id = document.get("source", "")
        content_hash = document.get("content_hash", "")

        # Update document
        self.documents[doc_id] = document
        self._save_document(doc_id)

        # Get existing elements and relationships for this document
        existing_elements = {e["element_id"]: e for e in self.get_document_elements(doc_id)}
        existing_relationships = {r["relationship_id"]: r
                                  for r in self.get_document_relationships(doc_id)}

        # Update elements - retain unchanged ones
        updated_element_ids = set()
        for element in elements:
            element_id = element["element_id"]
            updated_element_ids.add(element_id)

            if element_id in existing_elements:
                # Check if element has changed
                if self._has_element_changed(element, existing_elements[element_id]):
                    # Preserve the element_pk from the existing element
                    element["element_pk"] = existing_elements[element_id].get("element_pk")
                    self.elements[element_id] = element
                    self._save_element(element_id)
            else:
                # New element - assign an auto-increment element_pk
                element_pk = self.next_element_pk
                self.next_element_pk += 1
                element["element_pk"] = element_pk

                # Store mapping
                self.element_pks[element_id] = element_pk

                # Store element
                self.elements[element_id] = element
                self._save_element(element_id)

        # Remove elements that no longer exist
        for element_id in existing_elements:
            if element_id not in updated_element_ids:
                if element_id in self.elements:
                    # Get the element_pk for embedding deletion
                    element_pk = self.elements[element_id].get("element_pk")

                    # Delete element
                    del self.elements[element_id]
                    self._delete_element_file(element_id)

                    # Remove from element_pks mapping
                    if element_id in self.element_pks:
                        del self.element_pks[element_id]

                    # Also remove any embeddings
                    if element_pk in self.embeddings:
                        del self.embeddings[element_pk]
                        self._delete_embedding_file(element_pk)

        # Update relationships - similar approach
        updated_relationship_ids = set()
        for relationship in relationships:
            relationship_id = relationship["relationship_id"]
            updated_relationship_ids.add(relationship_id)

            if relationship_id in existing_relationships:
                # Check if relationship has changed
                if self._has_relationship_changed(relationship, existing_relationships[relationship_id]):
                    self.relationships[relationship_id] = relationship
                    self._save_relationship(relationship_id)
            else:
                # New relationship
                self.relationships[relationship_id] = relationship
                self._save_relationship(relationship_id)

        # Remove relationships that no longer exist
        for relationship_id in existing_relationships:
            if relationship_id not in updated_relationship_ids:
                if relationship_id in self.relationships:
                    del self.relationships[relationship_id]
                    self._delete_relationship_file(relationship_id)

        # Update processing history
        if source_id:
            self.update_processing_history(source_id, content_hash)

    def get_document(self, doc_id: str) -> Optional[Dict[str, Any]]:
        """Get document metadata by ID."""
        return self.documents.get(doc_id)

    def get_document_elements(self, doc_id: str) -> List[Dict[str, Any]]:
        """
        Get elements for a document by doc_id or source.
        Updated to match other implementations.
        """
        # First try to find document by doc_id
        document = self.documents.get(doc_id)

        # If not found by doc_id, try to find by source
        if not document:
            for doc in self.documents.values():
                if doc.get("source") == doc_id:
                    document = doc
                    doc_id = document["doc_id"]
                    break

        # If still not found, return empty list
        if not document:
            return []

        return [element for element in self.elements.values()
                if element.get("doc_id") == doc_id]

    def get_document_relationships(self, doc_id: str) -> List[Dict[str, Any]]:
        """Get relationships for a document."""
        # First get all element IDs for the document
        element_ids = {element["element_id"] for element in self.get_document_elements(doc_id)}

        # Find relationships involving these elements
        return [relationship for relationship in self.relationships.values()
                if relationship.get("source_id") in element_ids]

    def store_relationship(self, relationship: Dict[str, Any]) -> None:
        """
        Store a relationship between elements.

        Args:
            relationship: Relationship data with source_id, relationship_type, and target_reference
        """
        relationship_id = relationship["relationship_id"]
        self.relationships[relationship_id] = relationship
        self._save_relationship(relationship_id)
        logger.debug(f"Stored relationship {relationship_id}")

    def delete_relationships_for_element(self, element_id: str, relationship_type: str = None) -> None:
        """
        Delete relationships for an element.

        Args:
            element_id: Element ID
            relationship_type: Optional relationship type to filter by
        """
        # Find all relationships where this element is the source
        source_relationships = [rel_id for rel_id, rel in self.relationships.items()
                                if rel.get("source_id") == element_id and
                                (relationship_type is None or rel.get("relationship_type") == relationship_type)]

        # Find all relationships where this element is the target
        target_relationships = [rel_id for rel_id, rel in self.relationships.items()
                                if rel.get("target_reference") == element_id and
                                (relationship_type is None or rel.get("relationship_type") == relationship_type)]

        # Delete all matching relationships
        for rel_id in source_relationships + target_relationships:
            if rel_id in self.relationships:
                del self.relationships[rel_id]
                self._delete_relationship_file(rel_id)

        logger.debug(f"Deleted relationships for element {element_id}")

    def get_element(self, element_id_or_pk: Union[str, int]) -> Optional[Dict[str, Any]]:
        """
        Get element by ID or PK.
        Updated to handle either element_id or element_pk.

        Args:
            element_id_or_pk: Either the element_id (string) or element_pk (integer)
        """
        # Try to interpret as element_pk first
        try:
            element_pk = int(element_id_or_pk)
            # Find element with matching element_pk
            for element in self.elements.values():
                if element.get("element_pk") == element_pk:
                    return element
        except (ValueError, TypeError):
            # If not an integer, treat as element_id
            if element_id_or_pk in self.elements:
                return self.elements[element_id_or_pk]

        return None

    def find_documents(self, query: Dict[str, Any] = None, limit: int = 100) -> List[Dict[str, Any]]:
        """Find documents matching query."""
        if query is None:
            query = {}

        results = []

        for doc in self.documents.values():
            match = True

            for key, value in query.items():
                if key == "metadata":
                    # Check metadata fields
                    for meta_key, meta_value in value.items():
                        if meta_key not in doc.get("metadata", {}) or doc["metadata"][meta_key] != meta_value:
                            match = False
                            break
                elif key not in doc or doc[key] != value:
                    match = False
                    break

            if match:
                results.append(doc)
                if len(results) >= limit:
                    break

        return results

    def find_elements(self, query: Dict[str, Any] = None, limit: int = 100) -> List[Dict[str, Any]]:
        """Find elements matching query."""
        if query is None:
            query = {}

        results = []

        for element in self.elements.values():
            match = True

            for key, value in query.items():
                if key == "metadata":
                    # Check metadata fields
                    for meta_key, meta_value in value.items():
                        if meta_key not in element.get("metadata", {}) or element["metadata"][meta_key] != meta_value:
                            match = False
                            break
                elif key == "element_type" and isinstance(value, list):
                    # Handle list of allowed element types
                    if element.get("element_type") not in value:
                        match = False
                        break
                elif key == "doc_id" and isinstance(value, list):
                    # Handle list of document IDs to include
                    if element.get("doc_id") not in value:
                        match = False
                        break
                elif key == "exclude_doc_id" and isinstance(value, list):
                    # Handle list of document IDs to exclude
                    if element.get("doc_id") in value:
                        match = False
                        break
                elif key not in element or element[key] != value:
                    match = False
                    break

            if match:
                results.append(element)
                if len(results) >= limit:
                    break

        return results

    def search_elements_by_content(self, search_text: str, limit: int = 10) -> List[Dict[str, Any]]:
        """Search elements by content preview."""
        results = []
        search_text_lower = search_text.lower()

        for element in self.elements.values():
            content_preview = element.get("content_preview", "").lower()

            if search_text_lower in content_preview:
                results.append(element)

                if len(results) >= limit:
                    break

        return results

    def store_embedding(self, element_pk: int, embedding: List[float]) -> None:
        """
        Store embedding for an element.
        Already using element_pk as required.
        """
        # Verify element pk exists in some element
        found = False
        for element in self.elements.values():
            if element.get("element_pk") == element_pk:
                found = True
                break

        if not found:
            raise ValueError(f"Element pk not found: {element_pk}")

        self.embeddings[element_pk] = embedding
        self._save_embedding(element_pk)

    def get_embedding(self, element_pk: int) -> Optional[List[float]]:
        """
        Get embedding for an element.
        Already using element_pk as required.
        """
        return self.embeddings.get(element_pk)

    def search_by_embedding(self, query_embedding: List[float], limit: int = 10,
                            filter_criteria: Dict[str, Any] = None) -> List[Tuple[int, float]]:
        """
        Search elements by embedding similarity with optional filtering.
        Updated to return (element_pk, similarity) tuples for consistency.
        """
        import numpy as np

        query_embedding_np = np.array(query_embedding)
        results = []

        # Get all element_pks with embeddings
        element_pks_with_embeddings = list(self.embeddings.keys())

        # Build a dict of element_pk to element for easier lookup
        element_pk_to_element = {}
        for element in self.elements.values():
            if "element_pk" in element:
                element_pk_to_element[element["element_pk"]] = element

        # Apply filtering if specified
        if filter_criteria:
            filtered_element_pks = []
            for element_pk in element_pks_with_embeddings:
                # Get the associated element
                element = element_pk_to_element.get(element_pk)
                if not element:
                    continue

                # Check if element matches all criteria
                matches = True
                for key, value in filter_criteria.items():
                    if key == "element_type" and isinstance(value, list):
                        # Handle list of allowed element types
                        if element.get("element_type") not in value:
                            matches = False
                            break
                    elif key == "doc_id" and isinstance(value, list):
                        # Handle list of document IDs to include
                        if element.get("doc_id") not in value:
                            matches = False
                            break
                    elif key == "exclude_doc_id" and isinstance(value, list):
                        # Handle list of document IDs to exclude
                        if element.get("doc_id") in value:
                            matches = False
                            break
                    elif key == "exclude_doc_source" and isinstance(value, list):
                        # Handle list of document sources to exclude
                        doc_id = element.get("doc_id")
                        if doc_id:
                            doc = self.documents.get(doc_id)
                            if doc and doc.get("source") in value:
                                matches = False
                                break
                    else:
                        # Simple equality filter
                        if element.get(key) != value:
                            matches = False
                            break

                if matches:
                    filtered_element_pks.append(element_pk)
        else:
            filtered_element_pks = element_pks_with_embeddings

        # Calculate similarity for each filtered embedding
        for element_pk in filtered_element_pks:
            embedding = self.embeddings[element_pk]
            embedding_np = np.array(embedding)

            # Calculate cosine similarity
            dot_product = np.dot(query_embedding_np, embedding_np)
            norm1 = np.linalg.norm(query_embedding_np)
            norm2 = np.linalg.norm(embedding_np)

            if norm1 == 0 or norm2 == 0:
                similarity = 0.0
            else:
                similarity = float(dot_product / (norm1 * norm2))

            # Return element_pk instead of element_id for consistency
            results.append((element_pk, similarity))

        # Sort by similarity (highest first)
        results.sort(key=lambda x: x[1], reverse=True)

        return results[:limit]

    def delete_document(self, doc_id: str) -> bool:
        """Delete a document and all associated elements and relationships."""
        if doc_id not in self.documents:
            return False

        # Get elements to delete
        elements_to_delete = self.get_document_elements(doc_id)
        element_ids = {element["element_id"] for element in elements_to_delete}
        element_pks = {element.get("element_pk") for element in elements_to_delete if
                       element.get("element_pk") is not None}

        # Get relationships to delete
        relationships_to_delete = self.get_document_relationships(doc_id)
        relationship_ids = {rel["relationship_id"] for rel in relationships_to_delete}

        # Delete elements
        for element_id in element_ids:
            if element_id in self.elements:
                # Remove element_pk mapping
                if element_id in self.element_pks:
                    del self.element_pks[element_id]

                # Delete element
                del self.elements[element_id]
                self._delete_element_file(element_id)

        # Delete embeddings
        for element_pk in element_pks:
            if element_pk in self.embeddings:
                del self.embeddings[element_pk]
                self._delete_embedding_file(element_pk)

        # Delete relationships
        for relationship_id in relationship_ids:
            if relationship_id in self.relationships:
                del self.relationships[relationship_id]
                self._delete_relationship_file(relationship_id)

        # Delete document
        del self.documents[doc_id]
        self._delete_document_file(doc_id)

        return True

    def search_by_text(self, search_text: str, limit: int = 10,
                       filter_criteria: Dict[str, Any] = None) -> List[Tuple[int, float]]:
        """
        Search elements by semantic similarity to the provided text.
        Updated to return (element_pk, similarity) tuples for consistency.

        Args:
            search_text: Text to search for semantically
            limit: Maximum number of results
            filter_criteria: Optional dictionary with criteria to filter results

        Returns:
            List of (element_pk, similarity_score) tuples
        """
        try:
            if self.embedding_generator is None:
                from ..embeddings import get_embedding_generator
                self.embedding_generator = get_embedding_generator(config)

            # Generate embedding for the search text
            query_embedding = self.embedding_generator.generate(search_text)

            # Use the embedding to search, passing the filter criteria
            return self.search_by_embedding(query_embedding, limit, filter_criteria)

        except Exception as e:
            logger.error(f"Error in semantic search by text: {str(e)}")
            # Return empty list on error
            return []

    def _load_processing_history(self) -> None:
        """Load processing history from files."""
        history_files = glob.glob(os.path.join(self.storage_path, 'processing_history', '*.json'))

        for file_path in history_files:
            try:
                with open(file_path, 'r') as f:
                    history = json.load(f)

                # Use the safe filename as the key
                filename = os.path.basename(file_path)
                safe_id = os.path.splitext(filename)[0]
                self.processing_history[safe_id] = history

            except Exception as e:
                logger.error(f"Error loading processing history from {file_path}: {str(e)}")

    def _save_processing_history(self, safe_id: str) -> None:
        """Save processing history to file."""
        if safe_id not in self.processing_history:
            return

        file_path = os.path.join(self.storage_path, 'processing_history', f"{safe_id}.json")

        try:
            with open(file_path, 'w') as f:
                json.dump(self.processing_history[safe_id], f, indent=2)
        except Exception as e:
            logger.error(f"Error saving processing history to {file_path}: {str(e)}")

    @staticmethod
    def _get_safe_filename(source_id: str) -> str:
        """Convert source_id to a safe filename by hashing it."""
        import hashlib
        return hashlib.md5(source_id.encode('utf-8')).hexdigest()

    def _load_documents(self) -> None:
        """Load documents from files."""
        document_files = glob.glob(os.path.join(self.storage_path, 'documents', '*.json'))

        for file_path in document_files:
            try:
                with open(file_path, 'r') as f:
                    document = json.load(f)

                if "doc_id" in document:
                    self.documents[document["doc_id"]] = document
            except Exception as e:
                logger.error(f"Error loading document from {file_path}: {str(e)}")

    def _load_elements(self) -> None:
        """Load elements from files."""
        element_files = glob.glob(os.path.join(self.storage_path, 'elements', '*.json'))
        max_pk = 0  # Track the highest element_pk

        for file_path in element_files:
            try:
                with open(file_path, 'r') as f:
                    element = json.load(f)

                if "element_id" in element:
                    # Ensure element has element_pk
                    if "element_pk" not in element:
                        element["element_pk"] = self.next_element_pk
                        self.next_element_pk += 1

                    # Track the highest element_pk
                    max_pk = max(max_pk, element["element_pk"])

                    # Store element and mapping
                    self.elements[element["element_id"]] = element
                    self.element_pks[element["element_id"]] = element["element_pk"]
            except Exception as e:
                logger.error(f"Error loading element from {file_path}: {str(e)}")

        # Update next_element_pk to be one more than the highest seen
        self.next_element_pk = max_pk + 1

    def _load_relationships(self) -> None:
        """Load relationships from files."""
        relationship_files = glob.glob(os.path.join(self.storage_path, 'relationships', '*.json'))

        for file_path in relationship_files:
            try:
                with open(file_path, 'r') as f:
                    relationship = json.load(f)

                if "relationship_id" in relationship:
                    self.relationships[relationship["relationship_id"]] = relationship
            except Exception as e:
                logger.error(f"Error loading relationship from {file_path}: {str(e)}")

    def _load_embeddings(self) -> None:
        """Load embeddings from files."""
        import numpy as np

        embedding_files = glob.glob(os.path.join(self.storage_path, 'embeddings', '*.npy'))

        for file_path in embedding_files:
            try:
                # Extract element_pk from filename
                filename = os.path.basename(file_path)
                element_pk = int(os.path.splitext(filename)[0])

                # Load embedding
                embedding = np.load(file_path).tolist()
                self.embeddings[element_pk] = embedding
            except Exception as e:
                logger.error(f"Error loading embedding from {file_path}: {str(e)}")

    def _save_document(self, doc_id: str) -> None:
        """Save document to file."""
        if doc_id not in self.documents:
            return

        file_path = os.path.join(self.storage_path, 'documents', f"{doc_id}.json")

        try:
            with open(file_path, 'w') as f:
                json.dump(self.documents[doc_id], f, indent=2)
        except Exception as e:
            logger.error(f"Error saving document to {file_path}: {str(e)}")

    def _save_element(self, element_id: str) -> None:
        """Save element to file."""
        if element_id not in self.elements:
            return

        file_path = os.path.join(self.storage_path, 'elements', f"{element_id}.json")

        try:
            with open(file_path, 'w') as f:
                json.dump(self.elements[element_id], f, indent=2)
        except Exception as e:
            logger.error(f"Error saving element to {file_path}: {str(e)}")

    def _save_relationship(self, relationship_id: str) -> None:
        """Save relationship to file."""
        if relationship_id not in self.relationships:
            return

        file_path = os.path.join(self.storage_path, 'relationships', f"{relationship_id}.json")

        try:
            with open(file_path, 'w') as f:
                json.dump(self.relationships[relationship_id], f, indent=2)
        except Exception as e:
            logger.error(f"Error saving relationship to {file_path}: {str(e)}")

    def _save_embedding(self, element_pk: int) -> None:
        """Save embedding to file."""
        if element_pk not in self.embeddings:
            return

        import numpy as np

        file_path = os.path.join(self.storage_path, 'embeddings', f"{element_pk}.npy")

        try:
            np.save(file_path, np.array(self.embeddings[element_pk], dtype=np.float32))
        except Exception as e:
            logger.error(f"Error saving embedding to {file_path}: {str(e)}")

    def _delete_document_file(self, doc_id: str) -> None:
        """Delete document file."""
        file_path = os.path.join(self.storage_path, 'documents', f"{doc_id}.json")

        try:
            if os.path.exists(file_path):
                os.remove(file_path)
        except Exception as e:
            logger.error(f"Error deleting document file {file_path}: {str(e)}")

    def _delete_element_file(self, element_id: str) -> None:
        """Delete element file."""
        file_path = os.path.join(self.storage_path, 'elements', f"{element_id}.json")

        try:
            if os.path.exists(file_path):
                os.remove(file_path)
        except Exception as e:
            logger.error(f"Error deleting element file {file_path}: {str(e)}")

    def _delete_relationship_file(self, relationship_id: str) -> None:
        """Delete relationship file."""
        file_path = os.path.join(self.storage_path, 'relationships', f"{relationship_id}.json")

        try:
            if os.path.exists(file_path):
                os.remove(file_path)
        except Exception as e:
            logger.error(f"Error deleting relationship file {file_path}: {str(e)}")

    def _delete_embedding_file(self, element_pk: int) -> None:
        """Delete embedding file."""
        file_path = os.path.join(self.storage_path, 'embeddings', f"{element_pk}.npy")

        try:
            if os.path.exists(file_path):
                os.remove(file_path)
        except Exception as e:
            logger.error(f"Error deleting embedding file {file_path}: {str(e)}")

    @staticmethod
    def _has_element_changed(new_element: Dict[str, Any],
                             old_element: Dict[str, Any]) -> bool:
        """Check if element has changed."""
        # Check content hash first
        if new_element.get("content_hash") != old_element.get("content_hash"):
            return True

        # Check other fields
        for field in ["element_type", "parent_id", "content_preview", "content_location"]:
            if new_element.get(field) != old_element.get(field):
                return True

        # Check metadata
        new_metadata = new_element.get("metadata", {})
        old_metadata = old_element.get("metadata", {})

        if set(new_metadata.keys()) != set(old_metadata.keys()):
            return True

        for key, value in new_metadata.items():
            if old_metadata.get(key) != value:
                return True

        return False

    @staticmethod
    def _has_relationship_changed(new_rel: Dict[str, Any],
                                  old_rel: Dict[str, Any]) -> bool:
        """Check if relationship has changed."""
        for field in ["source_id", "relationship_type", "target_reference"]:
            if new_rel.get(field) != old_rel.get(field):
                return True

        # Check metadata
        new_metadata = new_rel.get("metadata", {})
        old_metadata = old_rel.get("metadata", {})

        if set(new_metadata.keys()) != set(old_metadata.keys()):
            return True

        for key, value in new_metadata.items():
            if old_metadata.get(key) != value:
                return True

        return False



import logging
import os


def configure_logging():
    # Get log level from an environment variable (default to INFO)
    log_level = os.environ.get("LOG_LEVEL", "INFO").upper()
    logging.basicConfig(
        level=getattr(logging, log_level, logging.INFO),  # Set global log level
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )



import uuid
from enum import Enum
from typing import Dict, Any, List

from .base import RelationshipDetector
from ..storage import ElementType


class RelationshipType(Enum):
    """Enumeration of structural relationship types."""

    HAS_TEMPORAL_INFORMATION = "has_temporal_information"
    TEMPORAL_RELATIONSHIP = "temporal_relationship"
    RELATED_TO = "related_to"
    DESCRIBED_BY = "described_by"
    DESCRIBES = "describes"
    NEXT_SIBLING = "next_sibling"
    PREVIOUS_SIBLING = "previous_sibling"
    LINK = "link"
    REFERENCED_BY = "referenced_by"
    SEMANTIC_SIMILARITY = "semantic_similarity"

    CONTAINS = "contains"
    CONTAINS_TABLE_HEADER = "contains_table_header"
    CONTAINS_LIST_ITEM = "contains_list_item"
    CONTAINS_TABLE_CELL = "contains_table_cell"
    CONTAINS_ARRAY_ITEM = "contains_array_item"  # New type for array items
    CONTAINS_ROW = "contains_row"
    CONTAINS_CELL = "contains_cell"
    CONTAINS_ITEM = "contains_item"
    CONTAINS_TEXT = "contains_text"  # Added to handle text content
    CONTAINS_TABLE_ROW = "contains_table_row"
    CONTAINS_NOTES = "contains_notes"

    CONTAINED_BY = "contained_by"

    @classmethod
    def is_parent_type(cls, rel_type: 'RelationshipType') -> bool:
        """Check if this is a parent relationship type."""
        return rel_type in [
            cls.CONTAINS,
            cls.CONTAINS_TABLE_HEADER,
            cls.CONTAINS_LIST_ITEM,
            cls.CONTAINS_TABLE_CELL,
            cls.CONTAINS_ARRAY_ITEM,
            cls.CONTAINS_ROW,
            cls.CONTAINS_CELL,
            cls.CONTAINS_ITEM,
            cls.CONTAINS_TEXT,
            cls.CONTAINS_TABLE_ROW,
            cls.CONTAINS_NOTES,
        ]

    @classmethod
    def is_child_type(cls, rel_type: 'RelationshipType') -> bool:
        """Check if this is a child relationship type."""
        return rel_type == cls.CONTAINED_BY

    @classmethod
    def is_sibling_type(cls, rel_type: 'RelationshipType') -> bool:
        """Check if this is a sibling relationship type."""
        return rel_type in [cls.NEXT_SIBLING, cls.PREVIOUS_SIBLING]

    @classmethod
    def is_semantic_type(cls, rel_type: 'RelationshipType') -> bool:
        """Check if this is a semantic relationship type."""
        return rel_type == cls.SEMANTIC_SIMILARITY

    @classmethod
    def is_link_type(cls, rel_type: 'RelationshipType') -> bool:
        """Check if this is a link relationship type."""
        return rel_type in [cls.LINK, cls.REFERENCED_BY]


class StructuralRelationshipDetector(RelationshipDetector):
    """Detector for structural relationships between elements."""

    def __init__(self, config: Dict[str, Any] = None):
        """
        Initialize the structural relationship detector.

        Args:
            config: Configuration dictionary
        """
        self.config = config or {}

    def detect_relationships(self, document: Dict[str, Any],
                             elements: List[Dict[str, Any]],
                             links: List[Dict[str, Any]] = None) -> List[Dict[str, Any]]:
        """Detect structural relationships between elements not found by parser.
        Generally, this is just creates sibling relationships between elements, or cleans up anything
        not handled by the parser.
        """
        relationships = []
        doc_id = document["doc_id"]

        # Create parent-child mapping
        parent_children = {}
        for element in elements:
            parent_id = element.get("parent_id")
            element_id = element["element_id"]

            if parent_id:
                if parent_id not in parent_children:
                    parent_children[parent_id] = []

                parent_children[parent_id].append(element_id)

        # Create sibling relationships
        for parent_id, children in parent_children.items():
            # Skip if only one child
            if len(children) <= 1:
                continue

            # Create relationships between consecutive siblings
            for i in range(len(children) - 1):
                prev_id = children[i]
                next_id = children[i + 1]

                # Create relationships
                relationship_id = self._generate_id("rel_")

                relationship = {
                    "relationship_id": relationship_id,
                    "doc_id": doc_id,
                    "source_id": prev_id,
                    "relationship_type": RelationshipType.NEXT_SIBLING.value,
                    "target_reference": next_id,
                    "metadata": {
                        "confidence": 1.0
                    }
                }

                relationships.append(relationship)

                # Create reverse relationship
                relationship_id = self._generate_id("rel_")

                relationship = {
                    "relationship_id": relationship_id,
                    "doc_id": doc_id,
                    "source_id": next_id,
                    "relationship_type": RelationshipType.PREVIOUS_SIBLING.value,
                    "target_reference": prev_id,
                    "metadata": {
                        "confidence": 1.0
                    }
                }

                relationships.append(relationship)

        return relationships

    @staticmethod
    def _get_section_elements(header: Dict[str, Any], _all_headers: List[Dict[str, Any]],
                              all_elements: List[Dict[str, Any]]) -> List[str]:
        """
        Get elements that belong to a header's section.

        A section includes all elements that:
        1. Come after this header
        2. Come before the next header of equal or higher level
        3. Are not headers themselves

        Args:
            header: Header element
            _all_headers: List of all headers
            all_elements: List of all elements

        Returns:
            List of element IDs in the section
        """
        header_id = header["element_id"]
        header_level = header.get("metadata", {}).get("level", 0)

        # Create list of elements in document order
        # This assumes elements are provided in document order
        element_ids = [e["element_id"] for e in all_elements]

        # Find index of this header
        try:
            header_index = element_ids.index(header_id)
        except ValueError:
            return []

        section_element_ids = []

        # Iterate through elements after this header
        for i in range(header_index + 1, len(element_ids)):
            element_id = element_ids[i]
            element = next((e for e in all_elements if e["element_id"] == element_id), None)

            if not element:
                continue

            # Stop at next header of equal or higher level
            if element.get("element_type") == ElementType.HEADER.value:
                element_level = element.get("metadata", {}).get("level", 0)
                if element_level <= header_level:
                    break

            # Add non-header elements to section
            if element.get("element_type") != ElementType.HEADER.value:
                section_element_ids.append(element_id)

        return section_element_ids

    @staticmethod
    def _generate_id(prefix: str = "") -> str:
        """
        Generate a unique ID.

        Args:
            prefix: Optional prefix for the ID

        Returns:
            Unique ID string
        """
        return f"{prefix}{uuid.uuid4()}"



"""
ServiceNow Content Source for the document pointer system.

This module provides integration with ServiceNow via its REST APIs.
"""

import json
import logging
import re
from typing import Dict, Any, List, Optional, TYPE_CHECKING, Tuple

import time

from .base import ContentSource

# Import types for type checking only - these won't be imported at runtime
if TYPE_CHECKING:
    import requests
    from requests import Session, Response
    from bs4 import BeautifulSoup

    # Define type aliases for type checking
    RequestsSessionType = Session
    RequestsResponseType = Response
    BeautifulSoupType = BeautifulSoup
else:
    # Runtime type aliases - use generic Python types
    RequestsSessionType = Any
    RequestsResponseType = Any
    BeautifulSoupType = Any

logger = logging.getLogger(__name__)

# Define global flags for availability - these will be set at runtime
REQUESTS_AVAILABLE = False
BS4_AVAILABLE = False

# Try to import requests conditionally
try:
    import requests

    REQUESTS_AVAILABLE = True
except ImportError:
    logger.warning("requests not available. Install with 'pip install requests' to use ServiceNow content source.")

# Try to import BeautifulSoup conditionally
try:
    from bs4 import BeautifulSoup

    BS4_AVAILABLE = True
except ImportError:
    logger.warning("beautifulsoup4 not available. Install with 'pip install beautifulsoup4' for improved HTML parsing.")


class ServiceNowContentSource(ContentSource):
    """Content source for ServiceNow."""

    def __init__(self, config: Dict[str, Any]):
        """
        Initialize the ServiceNow content source.

        Args:
            config: Configuration dictionary containing ServiceNow connection details
        """
        if not REQUESTS_AVAILABLE:
            raise ImportError("requests is required for ServiceNowContentSource but not available")

        super().__init__(config)
        self.base_url = config.get("base_url", "").rstrip('/')
        self.username = config.get("username", "")
        self.api_token = config.get("api_token", "")
        self.password = config.get("password", "")

        # Content type settings
        self.include_knowledge = config.get("include_knowledge", True)
        self.include_incidents = config.get("include_incidents", False)
        self.include_service_catalog = config.get("include_service_catalog", False)
        self.include_cmdb = config.get("include_cmdb", False)

        # Filter settings
        self.knowledge_query = config.get("knowledge_query", "")
        self.incident_query = config.get("incident_query", "")
        self.service_catalog_query = config.get("service_catalog_query", "")
        self.cmdb_query = config.get("cmdb_query", "")
        self.include_patterns = config.get("include_patterns", [])
        self.exclude_patterns = config.get("exclude_patterns", [])
        self.limit = config.get("limit", 100)

        # API paths
        self.table_api_path = "/api/now/table"
        self.knowledge_api_path = "/api/sn_km_api/knowledge"

        # Link following configuration
        self.max_link_depth = config.get("max_link_depth", 3)

        # Initialize session
        self.session: Optional[RequestsSessionType] = None
        try:
            self.session = requests.Session()

            # Set up authentication
            if self.username:
                if self.api_token:
                    # Use API token authentication
                    self.session.auth = (self.username, self.api_token)
                elif self.password:
                    # Use basic authentication
                    self.session.auth = (self.username, self.password)

            # Add common headers
            self.session.headers.update({
                "Accept": "application/json",
                "Content-Type": "application/json"
            })

            logger.debug(f"Successfully initialized session for ServiceNow: {self.get_safe_connection_string()}")
        except Exception as e:
            logger.error(f"Error initializing ServiceNow session: {str(e)}")
            raise

        # Cache for content
        self.content_cache = {}

    def get_safe_connection_string(self) -> str:
        """Return a safe version of the connection string with credentials masked."""
        if not self.base_url:
            return "<no base URL>"

        # Only show the base URL without credentials
        return self.base_url

    def fetch_document(self, source_id: str) -> Dict[str, Any]:
        """
        Fetch document content from ServiceNow.

        Args:
            source_id: Identifier for the content

        Returns:
            Dictionary containing document content and metadata

        Raises:
            ValueError: If ServiceNow is not configured or document not found
        """
        if not self.session:
            raise ValueError("ServiceNow not configured")

        logger.debug(f"Fetching ServiceNow content: {source_id}")

        try:
            # Parse the source ID to determine what type of content to fetch
            content_type, item_id = self._parse_source_id(source_id)

            # Handle different content types
            if content_type == "knowledge":
                return self._fetch_knowledge_article(item_id, source_id)
            elif content_type == "incident":
                return self._fetch_incident(item_id, source_id)
            elif content_type == "catalog_item":
                return self._fetch_catalog_item(item_id, source_id)
            elif content_type == "cmdb":
                return self._fetch_cmdb_item(item_id, source_id)
            else:
                raise ValueError(f"Unsupported content type: {content_type}")

        except ValueError:
            # Re-raise ValueError for not found or unsupported type
            raise
        except Exception as e:
            logger.error(f"Error fetching ServiceNow content {source_id}: {str(e)}")
            raise

    def list_documents(self) -> List[Dict[str, Any]]:
        """
        List available documents in ServiceNow.

        Returns:
            List of document identifiers and metadata

        Raises:
            ValueError: If ServiceNow is not configured
        """
        if not self.session:
            raise ValueError("ServiceNow not configured")

        logger.debug("Listing ServiceNow content")
        results = []

        try:
            # Fetch knowledge articles if enabled
            if self.include_knowledge:
                logger.debug("Fetching knowledge articles")
                knowledge_articles = self._list_knowledge_articles()
                results.extend(knowledge_articles)

                # Apply limits if needed
                if len(results) >= self.limit:
                    logger.debug(f"Reached limit of {self.limit} documents")
                    results = results[:self.limit]
                    return results

            # Fetch incidents if enabled
            if self.include_incidents:
                logger.debug("Fetching incidents")
                incidents = self._list_incidents()
                results.extend(incidents)

                # Apply limits if needed
                if len(results) >= self.limit:
                    logger.debug(f"Reached limit of {self.limit} documents")
                    results = results[:self.limit]
                    return results

            # Fetch service catalog items if enabled
            if self.include_service_catalog:
                logger.debug("Fetching service catalog items")
                catalog_items = self._list_catalog_items()
                results.extend(catalog_items)

                # Apply limits if needed
                if len(results) >= self.limit:
                    logger.debug(f"Reached limit of {self.limit} documents")
                    results = results[:self.limit]
                    return results

            # Fetch CMDB items if enabled
            if self.include_cmdb:
                logger.debug("Fetching CMDB items")
                cmdb_items = self._list_cmdb_items()
                results.extend(cmdb_items)

                # Apply limits if needed
                if len(results) >= self.limit:
                    logger.debug(f"Reached limit of {self.limit} documents")
                    results = results[:self.limit]
                    return results

            logger.info(f"Found {len(results)} ServiceNow documents")
            return results

        except Exception as e:
            logger.error(f"Error listing ServiceNow documents: {str(e)}")
            raise

    def has_changed(self, source_id: str, last_modified: Optional[float] = None) -> bool:
        """
        Check if a document has changed since last processing.

        Args:
            source_id: Identifier for the document
            last_modified: Timestamp of last known modification

        Returns:
            True if document has changed, False otherwise
        """
        if not self.session:
            # Can't determine changes without connection
            return True

        logger.debug(f"Checking if ServiceNow content has changed: {source_id}")

        try:
            # Parse the source ID
            content_type, item_id = self._parse_source_id(source_id)

            # If we have it in cache, check cache first
            if item_id in self.content_cache:
                cache_entry = self.content_cache[item_id]
                cache_modified = self._parse_servicenow_timestamp(cache_entry.get("last_modified", ""))

                if cache_modified and last_modified and cache_modified <= last_modified:
                    logger.debug(f"Content {item_id} unchanged according to cache")
                    return False

            # Make API request to check last modified date
            if content_type == "knowledge":
                api_url = f"{self.base_url}{self.knowledge_api_path}/articles/{item_id}"
                field_name = "sys_updated_on"
            else:
                # For other types, use the table API
                table_name = {
                    "incident": "incident",
                    "catalog_item": "sc_cat_item",
                    "cmdb": "cmdb_ci"
                }.get(content_type)

                api_url = f"{self.base_url}{self.table_api_path}/{table_name}/{item_id}"
                field_name = "sys_updated_on"

            # Make the request
            response = self.session.get(api_url)
            response.raise_for_status()

            # Get the response data
            data = response.json()

            # Extract last modified timestamp
            if content_type == "knowledge":
                current_modified = data.get("result", {}).get(field_name)
            else:
                current_modified = data.get("result", {}).get(field_name)

            # Parse the timestamp
            current_timestamp = self._parse_servicenow_timestamp(current_modified)

            # Compare timestamps
            if current_timestamp and last_modified:
                changed = current_timestamp > last_modified
                logger.debug(f"Content {item_id} changed: {changed}")
                return changed

            # If we can't determine based on timestamp, consider it changed
            return True

        except Exception as e:
            logger.error(f"Error checking changes for {source_id}: {str(e)}")
            return True  # Assume changed if there's an error

    def follow_links(self, content: str, source_id: str, current_depth: int = 0,
                     global_visited_docs=None) -> List[Dict[str, Any]]:
        """
        Extract and follow links in ServiceNow content.

        Args:
            content: Document content
            source_id: Identifier for the source document
            current_depth: Current depth of link following
            global_visited_docs: Global set of all visited document IDs

        Returns:
            List of linked documents
        """
        if not self.session:
            raise ValueError("ServiceNow not configured")

        # Check if we've reached the maximum link depth
        if current_depth >= self.max_link_depth:
            logger.debug(f"Max link depth {self.max_link_depth} reached for {source_id}")
            return []

        # Initialize global visited set if not provided
        if global_visited_docs is None:
            global_visited_docs = set()

        # Add current document to global visited set
        global_visited_docs.add(source_id)

        logger.debug(f"Following links in ServiceNow content {source_id} at depth {current_depth}")

        linked_docs = []

        # For ServiceNow content, we need to handle the different content types differently
        content_type, item_id = self._parse_source_id(source_id)

        # For HTML content (like in knowledge articles), extract links
        if content_type == "knowledge" and isinstance(content, str):
            try:
                if BS4_AVAILABLE:
                    soup = BeautifulSoup(content, 'html.parser')

                    # Find all anchor tags with href attributes
                    for a_tag in soup.find_all('a', href=True):
                        href = a_tag['href']

                        # Try to extract ServiceNow KB article ID from the href
                        kb_match = re.search(r'kb_view\.do\?sys_kb_id=([a-f0-9]+)', href)
                        if kb_match:
                            kb_id = kb_match.group(1)
                            linked_id = f"servicenow://{self.base_url}/knowledge/{kb_id}"

                            if linked_id not in global_visited_docs:
                                global_visited_docs.add(linked_id)
                                try:
                                    linked_doc = self.fetch_document(linked_id)
                                    linked_docs.append(linked_doc)
                                    logger.debug(f"Successfully fetched linked knowledge article: {kb_id}")

                                    # Recursively follow links if not at max depth
                                    if current_depth + 1 < self.max_link_depth:
                                        nested_docs = self.follow_links(
                                            linked_doc["content"],
                                            linked_doc["id"],
                                            current_depth + 1,
                                            global_visited_docs
                                        )
                                        linked_docs.extend(nested_docs)
                                except Exception as e:
                                    logger.warning(f"Error following knowledge article link {kb_id}: {str(e)}")

                        # Try to extract ServiceNow incident ID from the href
                        inc_match = re.search(r'incident\.do\?sys_id=([a-f0-9]+)', href)
                        if inc_match:
                            inc_id = inc_match.group(1)
                            linked_id = f"servicenow://{self.base_url}/incident/{inc_id}"

                            if linked_id not in global_visited_docs:
                                global_visited_docs.add(linked_id)
                                try:
                                    linked_doc = self.fetch_document(linked_id)
                                    linked_docs.append(linked_doc)
                                    logger.debug(f"Successfully fetched linked incident: {inc_id}")
                                except Exception as e:
                                    logger.warning(f"Error following incident link {inc_id}: {str(e)}")
                else:
                    logger.warning("BeautifulSoup is not available, skipping HTML link extraction")
            except Exception as e:
                logger.warning(f"Error parsing HTML for links: {str(e)}")

        # For JSON content (like incidents, catalog items), extract related records
        elif isinstance(content, dict):
            # Extract related incident from CMDB
            if content_type == "cmdb" and self.include_incidents:
                try:
                    # Get related incidents from the cmdb_ci_incident table
                    api_url = f"{self.base_url}{self.table_api_path}/cmdb_ci_incident"
                    params = {
                        "sysparm_query": f"cmdb_ci={item_id}",
                        "sysparm_limit": 10
                    }

                    response = self.session.get(api_url, params=params)
                    response.raise_for_status()

                    related_incidents = response.json().get("result", [])

                    for incident in related_incidents:
                        inc_id = incident.get("incident")
                        if inc_id:
                            linked_id = f"servicenow://{self.base_url}/incident/{inc_id}"

                            if linked_id not in global_visited_docs:
                                global_visited_docs.add(linked_id)
                                try:
                                    linked_doc = self.fetch_document(linked_id)
                                    linked_docs.append(linked_doc)
                                    logger.debug(f"Successfully fetched related incident: {inc_id}")
                                except Exception as e:
                                    logger.warning(f"Error following related incident {inc_id}: {str(e)}")

                except Exception as e:
                    logger.warning(f"Error fetching related incidents for CMDB item {item_id}: {str(e)}")

            # Extract CMDB CI from incident
            elif content_type == "incident" and self.include_cmdb:
                try:
                    # Get CMDB CI from the incident
                    cmdb_ci = content.get("cmdb_ci")

                    if cmdb_ci:
                        linked_id = f"servicenow://{self.base_url}/cmdb/{cmdb_ci}"

                        if linked_id not in global_visited_docs:
                            global_visited_docs.add(linked_id)
                            try:
                                linked_doc = self.fetch_document(linked_id)
                                linked_docs.append(linked_doc)
                                logger.debug(f"Successfully fetched related CMDB item: {cmdb_ci}")
                            except Exception as e:
                                logger.warning(f"Error following related CMDB item {cmdb_ci}: {str(e)}")

                except Exception as e:
                    logger.warning(f"Error processing CMDB CI for incident {item_id}: {str(e)}")

        logger.debug(f"Completed following links from {source_id}: found {len(linked_docs)} linked documents")
        return linked_docs

    def _fetch_knowledge_article(self, article_id: str, source_id: str) -> Dict[str, Any]:
        """
        Fetch a knowledge article from ServiceNow.

        Args:
            article_id: Knowledge article ID
            source_id: Original source ID

        Returns:
            Dictionary with content and metadata

        Raises:
            ValueError: If article not found
        """
        # Construct API URL for knowledge article
        api_url = f"{self.base_url}{self.knowledge_api_path}/articles/{article_id}"

        try:
            # Make API request
            response = self.session.get(api_url)
            response.raise_for_status()

            # Parse response
            article_data = response.json().get("result", {})

            if not article_data:
                raise ValueError(f"Knowledge article not found: {article_id}")

            # Extract content details
            title = article_data.get("short_description", "")
            article_number = article_data.get("number", "")
            html_content = article_data.get("content", "")
            sys_id = article_data.get("sys_id", "")
            created_on = article_data.get("sys_created_on", "")
            updated_on = article_data.get("sys_updated_on", "")

            # Clean up HTML content if available
            if html_content and BS4_AVAILABLE:
                try:
                    soup = BeautifulSoup(html_content, 'html.parser')

                    # Remove script and style elements
                    for script in soup(["script", "style"]):
                        script.extract()

                    # Get clean HTML
                    html_content = str(soup)
                except Exception as e:
                    logger.warning(f"Error cleaning HTML content: {str(e)}")

            # Construct metadata
            metadata = {
                "title": title,
                "number": article_number,
                "sys_id": sys_id,
                "created_on": created_on,
                "updated_on": updated_on,
                "type": "knowledge",
                "url": f"{self.base_url}/kb_view.do?sys_kb_id={sys_id}",
                "api_url": api_url,
                "content_type": "html"  # Explicitly mark as HTML content
            }

            # Generate content hash for change detection
            content_hash = self.get_content_hash(html_content)

            # Cache the content
            self.content_cache[article_id] = {
                "content": html_content,
                "metadata": metadata,
                "hash": content_hash,
                "last_modified": updated_on,
                "last_accessed": time.time()
            }

            return {
                "id": source_id,
                "content": html_content,
                "doc_type": "html",  # Explicitly mark as HTML document type
                "metadata": metadata,
                "content_hash": content_hash
            }
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                raise ValueError(f"Knowledge article not found: {article_id}")
            raise
        except Exception as e:
            logger.error(f"Error fetching knowledge article {article_id}: {str(e)}")
            raise

    def _fetch_incident(self, incident_id: str, source_id: str) -> Dict[str, Any]:
        """
        Fetch an incident from ServiceNow.

        Args:
            incident_id: Incident sys_id
            source_id: Original source ID

        Returns:
            Dictionary with content and metadata

        Raises:
            ValueError: If incident not found
        """
        # Construct API URL for incident
        api_url = f"{self.base_url}{self.table_api_path}/incident/{incident_id}"

        try:
            # Make API request
            response = self.session.get(api_url)
            response.raise_for_status()

            # Parse response
            incident_data = response.json().get("result", {})

            if not incident_data:
                raise ValueError(f"Incident not found: {incident_id}")

            # Should we include journal/work notes?
            try:
                # Get journal entries (comments and work notes)
                journal_url = f"{self.base_url}{self.table_api_path}/sys_journal_field"
                journal_params = {
                    "sysparm_query": f"element_id={incident_id}^ORDERBYDESCsys_created_on",
                    "sysparm_limit": 50  # Limit the number of entries to avoid huge payloads
                }

                journal_response = self.session.get(journal_url, params=journal_params)
                journal_response.raise_for_status()

                # Add journal entries to incident data
                incident_data["journal_entries"] = journal_response.json().get("result", [])
            except Exception as e:
                logger.warning(f"Error fetching journal entries for incident {incident_id}: {str(e)}")

            # Extract content details
            number = incident_data.get("number", "")
            short_description = incident_data.get("short_description", "")
            state = incident_data.get("state", "")
            priority = incident_data.get("priority", "")
            created_on = incident_data.get("sys_created_on", "")
            updated_on = incident_data.get("sys_updated_on", "")

            # Construct metadata
            metadata = {
                "number": number,
                "short_description": short_description,
                "state": state,
                "priority": priority,
                "created_on": created_on,
                "updated_on": updated_on,
                "type": "incident",
                "url": f"{self.base_url}/nav_to.do?uri=incident.do?sys_id={incident_id}",
                "api_url": api_url,
                "content_type": "json"  # Explicitly mark as JSON content
            }

            # Generate content hash for change detection
            content_hash = self.get_content_hash(json.dumps(incident_data))

            # Cache the content
            self.content_cache[incident_id] = {
                "content": incident_data,
                "metadata": metadata,
                "hash": content_hash,
                "last_modified": updated_on,
                "last_accessed": time.time()
            }

            return {
                "id": source_id,
                "content": incident_data,
                "doc_type": "json",  # Explicitly mark as JSON document type
                "metadata": metadata,
                "content_hash": content_hash
            }
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                raise ValueError(f"Incident not found: {incident_id}")
            raise
        except Exception as e:
            logger.error(f"Error fetching incident {incident_id}: {str(e)}")
            raise

    def _fetch_catalog_item(self, item_id: str, source_id: str) -> Dict[str, Any]:
        """
        Fetch a catalog item from ServiceNow.

        Args:
            item_id: Catalog item sys_id
            source_id: Original source ID

        Returns:
            Dictionary with content and metadata

        Raises:
            ValueError: If catalog item not found
        """
        # Construct API URL for catalog item
        api_url = f"{self.base_url}{self.table_api_path}/sc_cat_item/{item_id}"

        try:
            # Make API request
            response = self.session.get(api_url)
            response.raise_for_status()

            # Parse response
            item_data = response.json().get("result", {})

            if not item_data:
                raise ValueError(f"Catalog item not found: {item_id}")

            # Should we include variables?
            try:
                # Get variables for this catalog item
                variables_url = f"{self.base_url}{self.table_api_path}/item_option_new"
                variables_params = {
                    "sysparm_query": f"cat_item={item_id}",
                    "sysparm_limit": 100
                }

                variables_response = self.session.get(variables_url, params=variables_params)
                variables_response.raise_for_status()

                # Add variables to item data
                item_data["variables"] = variables_response.json().get("result", [])
            except Exception as e:
                logger.warning(f"Error fetching variables for catalog item {item_id}: {str(e)}")

            # Extract content details
            name = item_data.get("name", "")
            short_description = item_data.get("short_description", "")
            category = item_data.get("category", "")
            created_on = item_data.get("sys_created_on", "")
            updated_on = item_data.get("sys_updated_on", "")

            # Construct metadata
            metadata = {
                "name": name,
                "short_description": short_description,
                "category": category,
                "created_on": created_on,
                "updated_on": updated_on,
                "type": "catalog_item",
                "url": f"{self.base_url}/nav_to.do?uri=sc_cat_item.do?sys_id={item_id}",
                "api_url": api_url,
                "content_type": "json"  # Explicitly mark as JSON content
            }

            # Generate content hash for change detection
            content_hash = self.get_content_hash(json.dumps(item_data))

            # Cache the content
            self.content_cache[item_id] = {
                "content": item_data,
                "metadata": metadata,
                "hash": content_hash,
                "last_modified": updated_on,
                "last_accessed": time.time()
            }

            return {
                "id": source_id,
                "content": item_data,
                "doc_type": "json",  # Explicitly mark as JSON document type
                "metadata": metadata,
                "content_hash": content_hash
            }
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                raise ValueError(f"Catalog item not found: {item_id}")
            raise
        except Exception as e:
            logger.error(f"Error fetching catalog item {item_id}: {str(e)}")
            raise

    def _fetch_cmdb_item(self, ci_id: str, source_id: str) -> Dict[str, Any]:
        """
        Fetch a CMDB CI from ServiceNow.

        Args:
            ci_id: CMDB CI sys_id
            source_id: Original source ID

        Returns:
            Dictionary with content and metadata

        Raises:
            ValueError: If CMDB item not found
        """
        # Construct API URL for CMDB CI
        api_url = f"{self.base_url}{self.table_api_path}/cmdb_ci/{ci_id}"

        try:
            # Make API request
            response = self.session.get(api_url)
            response.raise_for_status()

            # Parse response
            ci_data = response.json().get("result", {})

            if not ci_data:
                raise ValueError(f"CMDB item not found: {ci_id}")

            # Should we include relationships?
            try:
                # Get relationships for this CI
                relations_url = f"{self.base_url}{self.table_api_path}/cmdb_rel_ci"
                relations_params = {
                    "sysparm_query": f"parent={ci_id}^ORchild={ci_id}",
                    "sysparm_limit": 100
                }

                relations_response = self.session.get(relations_url, params=relations_params)
                relations_response.raise_for_status()

                # Add relationships to CI data
                ci_data["relationships"] = relations_response.json().get("result", [])
            except Exception as e:
                logger.warning(f"Error fetching relationships for CMDB CI {ci_id}: {str(e)}")

            # Extract content details
            name = ci_data.get("name", "")
            sys_class_name = ci_data.get("sys_class_name", "")
            short_description = ci_data.get("short_description", "")
            created_on = ci_data.get("sys_created_on", "")
            updated_on = ci_data.get("sys_updated_on", "")

            # Construct metadata
            metadata = {
                "name": name,
                "sys_class_name": sys_class_name,
                "short_description": short_description,
                "created_on": created_on,
                "updated_on": updated_on,
                "type": "cmdb",
                "url": f"{self.base_url}/nav_to.do?uri=cmdb_ci.do?sys_id={ci_id}",
                "api_url": api_url,
                "content_type": "json"  # Explicitly mark as JSON content
            }

            # Generate content hash for change detection
            content_hash = self.get_content_hash(json.dumps(ci_data))

            # Cache the content
            self.content_cache[ci_id] = {
                "content": ci_data,
                "metadata": metadata,
                "hash": content_hash,
                "last_modified": updated_on,
                "last_accessed": time.time()
            }

            return {
                "id": source_id,
                "content": ci_data,
                "doc_type": "json",  # Explicitly mark as JSON document type
                "metadata": metadata,
                "content_hash": content_hash
            }
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                raise ValueError(f"CMDB item not found: {ci_id}")
            raise
        except Exception as e:
            logger.error(f"Error fetching CMDB item {ci_id}: {str(e)}")
            raise

    def _list_knowledge_articles(self) -> List[Dict[str, Any]]:
        """
        List knowledge articles in ServiceNow.

        Returns:
            List of knowledge article identifiers and metadata
        """
        articles = []

        try:
            # Construct API URL
            api_url = f"{self.base_url}{self.knowledge_api_path}/articles"

            # Set up parameters
            params = {
                "limit": self.limit
            }

            # Add query if specified
            if self.knowledge_query:
                params["query"] = self.knowledge_query

            # Make API request
            response = self.session.get(api_url, params=params)
            response.raise_for_status()

            # Parse response
            data = response.json()

            # Process results
            for article in data.get("result", []):
                article_id = article.get("sys_id", "")
                title = article.get("short_description", "")
                article_number = article.get("number", "")

                # Apply pattern filters if configured
                if not self._should_include_content(title, article_id):
                    logger.debug(f"Skipping excluded knowledge article: {title} ({article_id})")
                    continue

                # Create fully qualified source identifier
                qualified_source = f"servicenow://{self.base_url}/knowledge/{article_id}"

                # Get basic metadata
                created_on = article.get("sys_created_on", "")
                updated_on = article.get("sys_updated_on", "")

                articles.append({
                    "id": qualified_source,
                    "metadata": {
                        "title": title,
                        "number": article_number,
                        "sys_id": article_id,
                        "created_on": created_on,
                        "updated_on": updated_on,
                        "type": "knowledge",
                        "url": f"{self.base_url}/kb_view.do?sys_kb_id={article_id}",
                        "content_type": "html"  # Explicitly mark as HTML content
                    },
                    "doc_type": "html"  # Also set document type for consistency
                })

            logger.debug(f"Found {len(articles)} knowledge articles")
            return articles

        except Exception as e:
            logger.error(f"Error listing knowledge articles: {str(e)}")
            return []

    def _list_incidents(self) -> List[Dict[str, Any]]:
        """
        List incidents in ServiceNow.

        Returns:
            List of incident identifiers and metadata
        """
        incidents = []

        try:
            # Construct API URL
            api_url = f"{self.base_url}{self.table_api_path}/incident"

            # Set up parameters
            params = {
                "sysparm_limit": self.limit
            }

            # Add query if specified
            if self.incident_query:
                params["sysparm_query"] = self.incident_query

            # Make API request
            response = self.session.get(api_url, params=params)
            response.raise_for_status()

            # Parse response
            data = response.json()

            # Process results
            for incident in data.get("result", []):
                incident_id = incident.get("sys_id", "")
                number = incident.get("number", "")
                short_description = incident.get("short_description", "")

                # Apply pattern filters if configured
                if not self._should_include_content(short_description, incident_id):
                    logger.debug(f"Skipping excluded incident: {short_description} ({incident_id})")
                    continue

                # Create fully qualified source identifier
                qualified_source = f"servicenow://{self.base_url}/incident/{incident_id}"

                # Get basic metadata
                state = incident.get("state", "")
                priority = incident.get("priority", "")
                created_on = incident.get("sys_created_on", "")
                updated_on = incident.get("sys_updated_on", "")

                incidents.append({
                    "id": qualified_source,
                    "metadata": {
                        "number": number,
                        "short_description": short_description,
                        "state": state,
                        "priority": priority,
                        "created_on": created_on,
                        "updated_on": updated_on,
                        "type": "incident",
                        "url": f"{self.base_url}/nav_to.do?uri=incident.do?sys_id={incident_id}",
                        "content_type": "json"  # Explicitly mark as JSON content
                    },
                    "doc_type": "json"  # Also set document type for consistency
                })

            logger.debug(f"Found {len(incidents)} incidents")
            return incidents

        except Exception as e:
            logger.error(f"Error listing incidents: {str(e)}")
            return []

    def _list_catalog_items(self) -> List[Dict[str, Any]]:
        """
        List catalog items in ServiceNow.

        Returns:
            List of catalog item identifiers and metadata
        """
        catalog_items = []

        try:
            # Construct API URL
            api_url = f"{self.base_url}{self.table_api_path}/sc_cat_item"

            # Set up parameters
            params = {
                "sysparm_limit": self.limit
            }

            # Add query if specified
            if self.service_catalog_query:
                params["sysparm_query"] = self.service_catalog_query

            # Make API request
            response = self.session.get(api_url, params=params)
            response.raise_for_status()

            # Parse response
            data = response.json()

            # Process results
            for item in data.get("result", []):
                item_id = item.get("sys_id", "")
                name = item.get("name", "")
                short_description = item.get("short_description", "")

                # Apply pattern filters if configured
                if not self._should_include_content(name, item_id):
                    logger.debug(f"Skipping excluded catalog item: {name} ({item_id})")
                    continue

                # Create fully qualified source identifier
                qualified_source = f"servicenow://{self.base_url}/catalog_item/{item_id}"

                # Get basic metadata
                category = item.get("category", "")
                created_on = item.get("sys_created_on", "")
                updated_on = item.get("sys_updated_on", "")

                catalog_items.append({
                    "id": qualified_source,
                    "metadata": {
                        "name": name,
                        "short_description": short_description,
                        "category": category,
                        "created_on": created_on,
                        "updated_on": updated_on,
                        "type": "catalog_item",
                        "url": f"{self.base_url}/nav_to.do?uri=sc_cat_item.do?sys_id={item_id}",
                        "content_type": "json"  # Explicitly mark as JSON content
                    },
                    "doc_type": "json"  # Also set document type for consistency
                })

            logger.debug(f"Found {len(catalog_items)} catalog items")
            return catalog_items

        except Exception as e:
            logger.error(f"Error listing catalog items: {str(e)}")
            return []

    def _list_cmdb_items(self) -> List[Dict[str, Any]]:
        """
        List CMDB CIs in ServiceNow.

        Returns:
            List of CMDB CI identifiers and metadata
        """
        cmdb_items = []

        try:
            # Construct API URL
            api_url = f"{self.base_url}{self.table_api_path}/cmdb_ci"

            # Set up parameters
            params = {
                "sysparm_limit": self.limit
            }

            # Add query if specified
            if self.cmdb_query:
                params["sysparm_query"] = self.cmdb_query

            # Make API request
            response = self.session.get(api_url, params=params)
            response.raise_for_status()

            # Parse response
            data = response.json()

            # Process results
            for item in data.get("result", []):
                item_id = item.get("sys_id", "")
                name = item.get("name", "")
                sys_class_name = item.get("sys_class_name", "")

                # Apply pattern filters if configured
                if not self._should_include_content(name, item_id):
                    logger.debug(f"Skipping excluded CMDB CI: {name} ({item_id})")
                    continue

                # Create fully qualified source identifier
                qualified_source = f"servicenow://{self.base_url}/cmdb/{item_id}"

                # Get basic metadata
                short_description = item.get("short_description", "")
                created_on = item.get("sys_created_on", "")
                updated_on = item.get("sys_updated_on", "")

                cmdb_items.append({
                    "id": qualified_source,
                    "metadata": {
                        "name": name,
                        "sys_class_name": sys_class_name,
                        "short_description": short_description,
                        "created_on": created_on,
                        "updated_on": updated_on,
                        "type": "cmdb",
                        "url": f"{self.base_url}/nav_to.do?uri=cmdb_ci.do?sys_id={item_id}",
                        "content_type": "json"  # Explicitly mark as JSON content
                    },
                    "doc_type": "json"  # Also set document type for consistency
                })

            logger.debug(f"Found {len(cmdb_items)} CMDB CIs")
            return cmdb_items

        except Exception as e:
            logger.error(f"Error listing CMDB CIs: {str(e)}")
            return []

    def _should_include_content(self, title: str, content_id: str) -> bool:
        """
        Check if content should be included based on configured patterns.

        Args:
            title: Content title
            content_id: Content ID

        Returns:
            True if content should be included, False otherwise
        """
        # Check exclude patterns first
        for pattern in self.exclude_patterns:
            if re.search(pattern, title) or re.search(pattern, content_id):
                logger.debug(f"Content {content_id} excluded by pattern: {pattern}")
                return False

        # If no include patterns are configured, include everything not excluded
        if not self.include_patterns:
            return True

        # Check include patterns
        for pattern in self.include_patterns:
            if re.search(pattern, title) or re.search(pattern, content_id):
                logger.debug(f"Content {content_id} included by pattern: {pattern}")
                return True

        # If include patterns are configured and none matched, exclude
        return False

    @staticmethod
    def _parse_source_id(source_id: str) -> Tuple[str, str]:
        """
        Parse source ID to extract content type and item ID.

        Args:
            source_id: Source identifier

        Returns:
            Tuple of (content_type, item_id)
        """
        # Check if it's a fully qualified source ID
        # Pattern: servicenow://instance/content_type/item_id
        qualified_match = re.search(r'servicenow://([^/]+)/([^/]+)/([^/]+)', source_id)
        if qualified_match:
            # base_url = qualified_match.group(1)
            content_type = qualified_match.group(2)
            item_id = qualified_match.group(3)
            return content_type, item_id

        # If not fully qualified, try other formats

        # Check if it's just an ID (could be any content type)
        if re.match(r'^[a-f0-9]{32}$', source_id):
            # This is a ServiceNow sys_id, assume knowledge article by default
            return "knowledge", source_id

        # For knowledge articles in kb_view.do format
        kb_match = re.search(r'kb_view\.do\?sys_kb_id=([a-f0-9]+)', source_id)
        if kb_match:
            return "knowledge", kb_match.group(1)

        # For incidents in incident.do format
        inc_match = re.search(r'incident\.do\?sys_id=([a-f0-9]+)', source_id)
        if inc_match:
            return "incident", inc_match.group(1)

        # For catalog items in sc_cat_item.do format
        cat_match = re.search(r'sc_cat_item\.do\?sys_id=([a-f0-9]+)', source_id)
        if cat_match:
            return "catalog_item", cat_match.group(1)

        # For CMDB CIs in cmdb_ci.do format
        cmdb_match = re.search(r'cmdb_ci\.do\?sys_id=([a-f0-9]+)', source_id)
        if cmdb_match:
            return "cmdb", cmdb_match.group(1)

        # If no match, assume it's a knowledge article ID
        return "knowledge", source_id

    @staticmethod
    def _parse_servicenow_timestamp(timestamp: str) -> Optional[float]:
        """
        Parse ServiceNow timestamp into epoch time.

        Args:
            timestamp: ServiceNow timestamp string (format: yyyy-mm-dd hh:mm:ss)

        Returns:
            Timestamp as epoch time or None if parsing fails
        """
        if not timestamp:
            return None

        try:
            # ServiceNow uses a standard format: 2023-05-01 12:34:56
            from datetime import datetime

            dt = datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
            return dt.timestamp()
        except Exception:
            try:
                # Try parsing with dateutil for more flexibility
                import dateutil.parser

                dt = dateutil.parser.parse(timestamp)
                return dt.timestamp()
            except Exception:
                return None

    def __del__(self):
        """Close session when object is deleted."""
        if self.session:
            try:
                self.session.close()
                logger.debug("Closed ServiceNow session")
            except Exception as e:
                logger.warning(f"Error closing ServiceNow session: {str(e)}")



from dataclasses import field
from enum import Enum
from typing import Dict, Any, Optional, List

from pydantic import BaseModel


class RelationshipCategory(Enum):
    """Enumeration of relationship categories based on how they were created."""
    STRUCTURAL = "STRUCTURAL"  # Relationships derived from document structure
    EXPLICIT_LINK = "EXPLICIT"  # Explicit links found in the document
    SEMANTIC = "SEMANTIC"  # Semantic similarity relationships
    UNKNOWN = "UNKNOWN"  # Unknown or custom relationship types


# Helper functions for working with element collections


class ElementRelationship(BaseModel):
    """
    Class for representing relationships between elements.
    Designed to avoid name collisions with SQLAlchemy's Relationship class.
    Enriched with PK values for both source and target elements.
    """
    # Primary identifier
    relationship_id: str

    # Source element identifiers
    source_id: str

    # Relationship type
    relationship_type: str

    # Target element identifiers
    target_reference: str
    target_element_pk: Optional[int] = None
    target_element_type: Optional[str] = None
    target_content_preview: Optional[str] = None

    # Optional doc_id for the relationship
    doc_id: Optional[str] = None

    # Additional metadata
    metadata: Dict[str, Any] = field(default_factory=dict)

    # Direction flag - whether the specified element is the source (True) or target (False)
    # This is especially useful when retrieving relationships for a specific element
    is_source: bool = True
    source_element_pk: Optional[int] = None
    source_element_type: Optional[str] = None

    def __str__(self) -> str:
        """String representation of the relationship."""
        source_info = f"{self.source_id}({self.source_element_pk})" if self.source_element_pk else self.source_id
        target_info = f"{self.target_reference}({self.target_element_pk})" if self.target_element_pk else self.target_reference

        direction = "-->" if self.is_source else "<--"
        return f"{source_info} {direction}[{self.relationship_type}]{direction} {target_info}"

    def get_category(self) -> RelationshipCategory:
        """
        Determine the category of this relationship based on its type.

        Returns:
            RelationshipCategory enum value
        """
        # Structural relationship types from StructuralRelationshipDetector
        structural_types = [
            "next_sibling", "previous_sibling", "contains", "contained_by",
            "contains_row", "contains_cell", "contains_item"
        ]

        # Explicit link types from ExplicitLinkDetector
        explicit_link_types = ["link", "referenced_by"]

        # Semantic similarity types from SemanticRelationshipDetector
        semantic_types = ["semantic_similarity"]

        # Check relationship type
        rel_type = self.relationship_type.lower()

        if rel_type in structural_types:
            return RelationshipCategory.STRUCTURAL
        elif rel_type in explicit_link_types:
            return RelationshipCategory.EXPLICIT_LINK
        elif rel_type in semantic_types:
            return RelationshipCategory.SEMANTIC
        else:
            return RelationshipCategory.UNKNOWN

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary format for storage."""
        return {
            "relationship_id": self.relationship_id,
            "source_id": self.source_id,
            "source_element_pk": self.source_element_pk,
            "source_element_type": self.source_element_type,
            "relationship_type": self.relationship_type,
            "target_reference": self.target_reference,
            "target_element_pk": self.target_element_pk,
            "target_element_type": self.target_element_type,
            "target_content_preview": self.target_content_preview,
            "doc_id": self.doc_id,
            "metadata": self.metadata
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any], is_source: bool = True) -> 'ElementRelationship':
        """
        Create an ElementRelationship instance from a dictionary.

        Args:
            data: Dictionary containing relationship data
            is_source: Whether the specified element is the source (True) or target (False)

        Returns:
            ElementRelationship instance
        """
        return cls(
            relationship_id=data.get("relationship_id", ""),
            source_id=data.get("source_id", ""),
            source_element_pk=data.get("source_element_pk"),
            source_element_type=data.get("source_element_type"),
            relationship_type=data.get("relationship_type", ""),
            target_reference=data.get("target_reference", ""),
            target_element_pk=data.get("target_element_pk"),
            target_element_type=data.get("target_element_type"),
            doc_id=data.get("doc_id"),
            metadata=data.get("metadata", {}),
            is_source=is_source
        )

    def get_related_element_id(self) -> str:
        """
        Get the ID of the related element, regardless of direction.

        Returns:
            Element ID of the related element (source or target)
        """
        return self.target_reference if self.is_source else self.source_id

    def get_related_element_pk(self) -> Optional[int]:
        """
        Get the PK of the related element, regardless of direction.

        Returns:
            Element PK of the related element (source or target)
        """
        return self.target_element_pk if self.is_source else self.source_element_pk

    def is_bidirectional(self) -> bool:
        """Check if this is a bidirectional relationship based on type."""
        # Semantic similarity is bidirectional
        if self.get_category() == RelationshipCategory.SEMANTIC:
            return True

        # Some specific structural relationships are bidirectional
        bidirectional_types = [
            "semantic_similarity", "synonym", "related",
            "similar", "equivalent"
        ]
        return self.relationship_type.lower() in bidirectional_types

    def get_confidence(self) -> float:
        """Get the confidence score for this relationship."""
        return self.metadata.get("confidence", 0.0)

    def get_similarity(self) -> float:
        """Get the similarity score for semantic relationships."""
        if self.get_category() == RelationshipCategory.SEMANTIC:
            return self.metadata.get("similarity", 0.0)
        return 0.0


# Helper functions for working with relationship collections

def get_structural_relationships(relationships: List[ElementRelationship]) -> List[ElementRelationship]:
    """
    Filter relationships to get only structural relationships.

    Args:
        relationships: List of ElementRelationship objects

    Returns:
        List of structural relationships
    """
    return [r for r in relationships if r.get_category() == RelationshipCategory.STRUCTURAL]


def get_explicit_links(relationships: List[ElementRelationship]) -> List[ElementRelationship]:
    """
    Filter relationships to get only explicit links.

    Args:
        relationships: List of ElementRelationship objects

    Returns:
        List of explicit link relationships
    """
    return [r for r in relationships if r.get_category() == RelationshipCategory.EXPLICIT_LINK]


def get_semantic_relationships(relationships: List[ElementRelationship]) -> List[ElementRelationship]:
    """
    Filter relationships to get only semantic similarity relationships.

    Args:
        relationships: List of ElementRelationship objects

    Returns:
        List of semantic similarity relationships
    """
    return [r for r in relationships if r.get_category() == RelationshipCategory.SEMANTIC]


def get_container_relationships(relationships: List[ElementRelationship]) -> List[ElementRelationship]:
    """
    Get container relationships (contains, contains_row, contains_cell, contains_item).

    Args:
        relationships: List of ElementRelationship objects

    Returns:
        List of container relationships
    """
    container_types = ["contains", "contains_row", "contains_cell", "contains_item"]
    return [r for r in relationships if r.relationship_type in container_types]


def get_sibling_relationships(relationships: List[ElementRelationship]) -> List[ElementRelationship]:
    """
    Get sibling relationships (next_sibling, previous_sibling).

    Args:
        relationships: List of ElementRelationship objects

    Returns:
        List of sibling relationships
    """
    sibling_types = ["next_sibling", "previous_sibling"]
    return [r for r in relationships if r.relationship_type in sibling_types]


def sort_relationships_by_confidence(relationships: List[ElementRelationship]) -> List[ElementRelationship]:
    """
    Sort relationships by confidence (highest first).

    Args:
        relationships: List of ElementRelationship objects

    Returns:
        Sorted list of relationships
    """
    return sorted(relationships, key=lambda r: r.get_confidence(), reverse=True)


def sort_semantic_relationships_by_similarity(relationships: List[ElementRelationship]) -> List[ElementRelationship]:
    """
    Sort semantic relationships by similarity (highest first).

    Args:
        relationships: List of ElementRelationship objects

    Returns:
        Sorted list of semantic relationships
    """
    semantic_rels = get_semantic_relationships(relationships)
    return sorted(semantic_rels, key=lambda r: r.get_similarity(), reverse=True)



"""
Utility functions for document parsing.

This module provides helper functions for detecting content types and routing documents
to the appropriate parser.
"""

import logging
import re
from typing import Dict, Any

logger = logging.getLogger(__name__)


def detect_content_type(content: str, metadata: Dict[str, Any] = None) -> str:
    """
    Detect content type from content and metadata.

    Args:
        content: Document content
        metadata: Optional metadata that might contain content type information

    Returns:
        Content type: 'markdown', 'html', or 'text'
    """
    metadata = metadata or {}

    # Check metadata first
    content_type = metadata.get("content_type", "")
    if content_type:
        if "markdown" in content_type.lower() or "md" in content_type.lower():
            return "markdown"
        elif "html" in content_type.lower() or "xhtml" in content_type.lower():
            return "html"

    # Check content patterns
    if content.strip().startswith(('<!DOCTYPE html>', '<html', '<?xml')):
        return "html"

    # Count markdown-specific patterns
    md_patterns = 0
    if re.search(r'^#{1,6}\s+', content, re.MULTILINE):  # Headers
        md_patterns += 1
    if re.search(r'^-\s+', content, re.MULTILINE):  # List items
        md_patterns += 1
    if re.search(r'\[.+?\]\(.+?\)', content):  # Links
        md_patterns += 1
    if re.search(r'^```', content, re.MULTILINE):  # Code blocks
        md_patterns += 1

    if md_patterns >= 2:
        return "markdown"

    # Count HTML-specific patterns
    html_patterns = 0
    if re.search(r'<[a-z]+[^>]*>', content):  # Opening tags
        html_patterns += 1
    if re.search(r'</[a-z]+>', content):  # Closing tags
        html_patterns += 1
    if re.search(r'<[a-z]+[^>]*/>', content):  # Self-closing tags
        html_patterns += 1

    if html_patterns >= 2:
        return "html"

    # Default to text
    return "text"


def extract_url_links(content: str, element_id: str) -> list:
    """
    Extract URL links from plain text.

    Args:
        content: Text content
        element_id: Source element ID

    Returns:
        List of link objects
    """
    links = []

    # Simple URL pattern for plain text
    url_pattern = r'https?://[^\s()<>]+(?:\([\w\d]+\)|(?:[^,.;:`!()\[\]{}<>"\'\s]|/))'

    matches = re.findall(url_pattern, content)

    for url in matches:
        links.append({
            "source_id": element_id,
            "link_text": url,  # Use the URL as the text for plain text links
            "link_target": url,
            "link_type": "url"
        })

    return links



"""
MongoDB adapter module for the document pointer system.

This module provides an adapter to retrieve content from MongoDB sources.
"""

import json
import logging
from datetime import datetime
from typing import Dict, Any, Optional

from .base import ContentSourceAdapter
from ..document_parser.document_type_detector import DocumentTypeDetector

logger = logging.getLogger(__name__)

# Try to import pymongo, but don't fail if not available
try:
    import pymongo
    from pymongo import MongoClient
    from bson import ObjectId, json_util

    PYMONGO_AVAILABLE = True
except ImportError:
    MongoClient = None
    ObjectId = None
    json_util = None
    PYMONGO_AVAILABLE = False
    logger.warning("pymongo not available. Install with 'pip install pymongo' to use MongoDB adapter.")


class MongoDBAdapter(ContentSourceAdapter):
    """Adapter for MongoDB content."""

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """Initialize the MongoDB adapter."""
        if not PYMONGO_AVAILABLE:
            raise ImportError("pymongo is required for MongoDB adapter")

        super().__init__(config)
        self.config = config or {}

        # MongoDB connection settings
        self.default_connection_string = self.config.get("connection_string")
        self.default_database = self.config.get("database")
        self.default_collection = self.config.get("collection")
        self.connection_timeout = self.config.get("connection_timeout", 30000)
        self.server_selection_timeout = self.config.get("server_selection_timeout", 30000)

        # Auth settings if not in connection string
        self.username = self.config.get("username")
        self.password = self.config.get("password")
        self.auth_source = self.config.get("auth_source", "admin")
        self.auth_mechanism = self.config.get("auth_mechanism", "SCRAM-SHA-256")

        # Cached connections
        self.clients = {}

        # Content cache
        self.content_cache = {}
        self.metadata_cache = {}

    def get_content(self, location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get content from MongoDB.

        Args:
            location_data: Location data with MongoDB connection info

        Returns:
            Dictionary with content and metadata

        Raises:
            ValueError: If MongoDB document cannot be retrieved
        """
        # Parse location data
        source = location_data.get("source", "")
        if not source.startswith("mongodb://"):
            raise ValueError(f"Invalid MongoDB source: {source}")

        # Cache key using the full source string
        cache_key = source
        if cache_key in self.content_cache:
            return self.content_cache[cache_key]

        # Parse MongoDB URI and extract components
        parsed_data = self._parse_mongodb_uri(source)

        # Get client for this connection
        client = self._get_client(parsed_data["connection_string"])

        try:
            # Access database and collection
            db = client[parsed_data["database"]]
            collection = db[parsed_data["collection"]]

            # Prepare query for finding the document
            query = self._build_query(parsed_data)

            # Execute query
            document = collection.find_one(query)

            if not document:
                raise ValueError(f"Document not found for query: {query}")

            # Extract field if specified
            content, field_path = self._extract_field(document, parsed_data)

            # Determine content type
            if isinstance(content, dict) or isinstance(content, list):
                content_type = "json"
                str_content = json.dumps(content, default=json_util.default, indent=2)
            else:
                # Try to detect content type from the content itself
                str_content = str(content) if not isinstance(content, str) else content
                content_type = DocumentTypeDetector.detect_from_content(
                    str_content,
                    metadata={"field_path": field_path}
                )

            # Prepare metadata
            metadata = {
                "connection_string_masked": self._mask_connection_string(parsed_data["connection_string"]),
                "database": parsed_data["database"],
                "collection": parsed_data["collection"],
                "document_id": str(document.get("_id")),
                "document_size": len(json.dumps(document, default=json_util.default)),
                "field_path": field_path if field_path else None,
                "query": str(query),
                "timestamp": datetime.now().isoformat()
            }

            # Add field type information
            if field_path:
                field_type = type(content).__name__
                metadata["field_type"] = field_type

            # Create result
            result = {
                "content": str_content if not isinstance(content, (bytes, bytearray)) else content,
                "content_type": content_type,
                "metadata": metadata
            }

            # Cache the result
            self.content_cache[cache_key] = result
            self.metadata_cache[cache_key] = metadata

            return result

        except Exception as e:
            logger.error(f"Error retrieving MongoDB content: {str(e)}")
            raise ValueError(f"Error retrieving MongoDB content: {str(e)}")

    def supports_location(self, location_data: Dict[str, Any]) -> bool:
        """
        Check if this adapter supports the location.

        Args:
            location_data: Content location data

        Returns:
            True if supported, False otherwise
        """
        if not PYMONGO_AVAILABLE:
            return False

        source = location_data.get("source", "")
        # Source must be a MongoDB URI
        return source.startswith("mongodb://")

    def get_binary_content(self, location_data: Dict[str, Any]) -> bytes:
        """
        Get the MongoDB document as binary data.

        Args:
            location_data: Location data

        Returns:
            Binary content

        Raises:
            ValueError: If document cannot be retrieved as binary
        """
        # First try to get the content normally
        content_data = self.get_content(location_data)
        content = content_data.get("content")

        # If content is already binary, return it
        if isinstance(content, (bytes, bytearray)):
            return content

        # Otherwise, convert to JSON and then to bytes
        if isinstance(content, (dict, list)):
            return json.dumps(content, default=json_util.default).encode('utf-8')

        # Convert string content to bytes
        if isinstance(content, str):
            return content.encode('utf-8')

        # For other types, convert to string first
        return str(content).encode('utf-8')

    def get_metadata(self, location_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get metadata about the MongoDB document without retrieving the full content.

        Args:
            location_data: Location data

        Returns:
            Dictionary with metadata

        Raises:
            ValueError: If metadata cannot be retrieved
        """
        source = location_data.get("source", "")

        # Check cache first
        if source in self.metadata_cache:
            return self.metadata_cache[source]

        # Parse MongoDB URI and extract components
        parsed_data = self._parse_mongodb_uri(source)

        # Get client for this connection
        client = self._get_client(parsed_data["connection_string"])

        try:
            # Access database and collection
            db = client[parsed_data["database"]]
            collection = db[parsed_data["collection"]]

            # Prepare query for finding the document
            query = self._build_query(parsed_data)

            # Execute projection query to just get basic info without the full content
            projection = {"_id": 1}
            if parsed_data.get("field_path"):
                # Add field existence check
                field_parts = parsed_data["field_path"].split(".")
                field_to_check = field_parts[0]
                projection[field_to_check] = 1

            document = collection.find_one(query, projection)

            if not document:
                raise ValueError(f"Document not found for query: {query}")

            # Prepare metadata
            metadata = {
                "connection_string_masked": self._mask_connection_string(parsed_data["connection_string"]),
                "database": parsed_data["database"],
                "collection": parsed_data["collection"],
                "document_id": str(document.get("_id")),
                "field_path": parsed_data.get("field_path"),
                "query": str(query),
                "timestamp": datetime.now().isoformat()
            }

            # Get collection stats
            try:
                stats = db.command("collstats", parsed_data["collection"])
                metadata["collection_size"] = stats.get("size")
                metadata["document_count"] = stats.get("count")
                metadata["avg_document_size"] = stats.get("avgObjSize")
            except Exception:
                # If stats command fails, just continue without this info
                pass

            # Cache the metadata
            self.metadata_cache[source] = metadata

            return metadata

        except Exception as e:
            logger.error(f"Error retrieving MongoDB metadata: {str(e)}")
            raise ValueError(f"Error retrieving MongoDB metadata: {str(e)}")

    def resolve_uri(self, uri: str) -> Dict[str, Any]:
        """
        Parse a MongoDB URI into location data.

        Args:
            uri: MongoDB URI string

        Returns:
            Dictionary with parsed location data

        Raises:
            ValueError: If URI cannot be parsed
        """
        if not uri.startswith("mongodb://"):
            raise ValueError(f"Not a MongoDB URI: {uri}")

        # Parse URI into components
        parsed_data = self._parse_mongodb_uri(uri)

        # Build location data
        location_data = {
            "source": uri,
            "connection_string": parsed_data["connection_string"],
            "database": parsed_data["database"],
            "collection": parsed_data["collection"]
        }

        # Add document_id if available
        if "document_id" in parsed_data:
            location_data["document_id"] = parsed_data["document_id"]

        # Add field_path if available
        if "field_path" in parsed_data:
            location_data["field_path"] = parsed_data["field_path"]

        return location_data

    def cleanup(self):
        """
        Clean up MongoDB connections.

        This method should be called when the adapter is no longer needed.
        """
        for conn_id, client in self.clients.items():
            try:
                client.close()
                logger.debug(f"Closed MongoDB connection: {conn_id}")
            except Exception as e:
                logger.warning(f"Error closing MongoDB connection {conn_id}: {str(e)}")

        # Clear caches and connections
        self.clients = {}
        self.content_cache = {}
        self.metadata_cache = {}

    @staticmethod
    def _parse_mongodb_uri(uri: str) -> Dict[str, Any]:
        """
        Parse MongoDB URI into components.

        Expected format: mongodb://[username:password@]host[:port]/database/collection[/document_id][/field_path]

        Args:
            uri: MongoDB URI string

        Returns:
            Dictionary with parsed components
        """
        # Split off the connection string part
        if not uri.startswith("mongodb://"):
            raise ValueError(f"Invalid MongoDB URI: {uri}")

        # Find the first / after mongodb://
        base_uri_parts = uri.split("/")

        # We need at least mongodb://host/database/collection
        if len(base_uri_parts) < 5:
            raise ValueError(f"Invalid MongoDB URI format. Expected mongodb://host/database/collection: {uri}")

        # Extract connection string (mongodb://host)
        connection_string = f"{base_uri_parts[0]}//{base_uri_parts[2]}"

        # Extract database and collection
        database = base_uri_parts[3]
        collection = base_uri_parts[4]

        # Create result with essential parts
        result = {
            "connection_string": connection_string,
            "database": database,
            "collection": collection
        }

        # Check for document_id and field_path
        if len(base_uri_parts) > 5:
            result["document_id"] = base_uri_parts[5]

            # Check for field path
            if len(base_uri_parts) > 6:
                # Rejoin remaining parts as the field path
                result["field_path"] = "/".join(base_uri_parts[6:])

        return result

    def _get_client(self, connection_string: str) -> MongoClient:
        """
        Get or create MongoDB client for the given connection string.

        Args:
            connection_string: MongoDB connection string

        Returns:
            MongoDB client

        Raises:
            ValueError: If connection cannot be established
        """
        # Use a masked version as the cache key
        cache_key = self._mask_connection_string(connection_string)

        # Check if client exists in cache
        if cache_key in self.clients:
            return self.clients[cache_key]

        # Create client options
        client_options = {
            "connectTimeoutMS": self.connection_timeout,
            "serverSelectionTimeoutMS": self.server_selection_timeout
        }

        # Add authentication if needed and not in connection string
        if self.username and self.password and "@" not in connection_string:
            client_options["username"] = self.username
            client_options["password"] = self.password
            client_options["authSource"] = self.auth_source
            client_options["authMechanism"] = self.auth_mechanism

        try:
            # Create MongoDB client
            client = MongoClient(connection_string, **client_options)

            # Test connection with a ping
            client.admin.command("ping")

            # Cache client
            self.clients[cache_key] = client

            return client

        except Exception as e:
            logger.error(f"Error connecting to MongoDB: {str(e)}")
            raise ValueError(f"Cannot connect to MongoDB: {str(e)}")

    @staticmethod
    def _build_query(parsed_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Build MongoDB query from parsed URI data.

        Args:
            parsed_data: Parsed MongoDB URI data

        Returns:
            MongoDB query document
        """
        query = {}

        # Add document ID if specified
        if "document_id" in parsed_data:
            doc_id = parsed_data["document_id"]

            # Try to convert to ObjectId if possible
            if ObjectId.is_valid(doc_id):
                query["_id"] = ObjectId(doc_id)
            else:
                # Use as string ID
                query["_id"] = doc_id

        return query

    @staticmethod
    def _extract_field(document: Dict[str, Any], parsed_data: Dict[str, Any]) -> tuple:
        """
        Extract field from document using dot notation.

        Args:
            document: MongoDB document
            parsed_data: Parsed MongoDB URI data

        Returns:
            Tuple of (field_value, field_path)
        """
        field_path = parsed_data.get("field_path")

        if not field_path:
            # Return full document if no field specified
            return document, None

        # Handle array indexing and nested fields
        parts = field_path.replace('[', '.').replace(']', '').split('.')

        current = document
        for part in parts:
            if not part:  # Skip empty parts
                continue

            # Handle numeric indices for arrays
            if part.isdigit() and isinstance(current, list):
                index = int(part)
                if 0 <= index < len(current):
                    current = current[index]
                else:
                    raise ValueError(f"Array index out of range: {part} in {field_path}")
            # Handle dictionary keys
            elif isinstance(current, dict) and part in current:
                current = current[part]
            else:
                raise ValueError(f"Field not found: {part} in {field_path}")

        return current, field_path

    @staticmethod
    def _mask_connection_string(connection_string: str) -> str:
        """
        Mask sensitive information in connection string for logging.

        Args:
            connection_string: MongoDB connection string

        Returns:
            Masked connection string
        """
        # Check if connection string contains credentials
        if "@" in connection_string:
            # Extract username:password part
            parts = connection_string.split("@")
            prefix = parts[0]
            suffix = "@".join(parts[1:])

            # Find where credentials are
            cred_parts = prefix.split("//")
            protocol = cred_parts[0] + "//"
            # credentials = cred_parts[1]

            # Mask credentials
            return f"{protocol}****:****@{suffix}"

        return connection_string



"""
Content Source Module for the document pointer system.
This module contains adapters for different content sources such as:
- Markdown files
- Database blobs
- Web URLs
- Confluence
- JIRA
- S3
- ServiceNow
- MongoDB
- SharePoint
- Google Drive
"""
import logging
from typing import Dict, Any

from .base import ContentSource
from .confluence import ConfluenceContentSource
from .database import DatabaseContentSource
from .file import FileContentSource
from .google_drive import GoogleDriveContentSource
from .jira import JiraContentSource
from .mongodb import MongoDBContentSource
from .s3 import S3ContentSource
from .servicenow import ServiceNowContentSource
from .sharepoint import SharePointContentSource
from .web import WebContentSource

logger = logging.getLogger(__name__)

def get_content_source(source_config: Dict[str, Any]) -> ContentSource:
    """
    Factory function to create appropriate content source from config.

    Args:
        source_config: Content source configuration

    Returns:
        ContentSource instance

    Raises:
        ValueError: If source type is not supported
    """
    source_type = source_config.get("type")

    if source_type == "file":
        return FileContentSource(source_config)
    elif source_type == "database":
        return DatabaseContentSource(source_config)
    elif source_type == "web":
        return WebContentSource(source_config)
    elif source_type == "confluence":
        return ConfluenceContentSource(source_config)
    elif source_type == "jira":
        return JiraContentSource(source_config)
    elif source_type == "s3":
        return S3ContentSource(source_config)
    elif source_type == "servicenow":
        return ServiceNowContentSource(source_config)
    elif source_type == "mongodb":
        return MongoDBContentSource(source_config)
    elif source_type == "sharepoint":
        return SharePointContentSource(source_config)
    elif source_type == "google_drive":
        return GoogleDriveContentSource(source_config)
    else:
        raise ValueError(f"Unsupported content source type: {source_type}")
